import bonobo


def split_one(line):
    return dict(zip(("name", "address"), line.split(', ', 1)))


graph = bonobo.Graph(
    bonobo.FileReader('coffeeshops.txt'),
    split_one,
    bonobo.JsonWriter('coffeeshops.json', fs='fs.output'),
)


def get_services():
    return {
        'fs': bonobo.open_examples_fs('datasets'),
        'fs.output': bonobo.open_fs(),
    }


if __name__ == '__main__':
    bonobo.run(graph, services=get_services())



"""
Extracts a list of fablabs in the world, restrict to the ones in france, then format it both for a nice console output
and a flat txt file.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "ODS()" -> "normalize" -> "filter_france" -> "Tee()" -> "JsonWriter()";
    }

"""

import json

import bonobo
from bonobo import examples
from bonobo.contrib.opendatasoft import OpenDataSoftAPI
from bonobo.examples.datasets.services import get_services

try:
    import pycountry
except ImportError as exc:
    raise ImportError(
        'You must install package "pycountry" to run this example.'
    ) from exc

API_DATASET = 'fablabs@public-us'
ROWS = 100


def _getlink(x):
    return x.get('url', None)


def normalize(row):
    result = {
        **row,
        'links': list(filter(None, map(_getlink, json.loads(row.get('links'))))),
        'country': pycountry.countries.get(alpha_2=row.get('country_code', '').upper()).name,
    }
    return result


def get_graph(graph=None, *, _limit=(), _print=()):
    graph = graph or bonobo.Graph()
    graph.add_chain(
        OpenDataSoftAPI(dataset=API_DATASET),
        *_limit,
        normalize,
        bonobo.UnpackItems(0),
        *_print,
        bonobo.JsonWriter(path='fablabs.json'),
    )
    return graph


if __name__ == '__main__':
    parser = examples.get_argument_parser()

    with bonobo.parse_args(parser) as options:
        bonobo.run(
            get_graph(**examples.get_graph_options(options)),
            services=get_services()
        )



import os

# https://developers.google.com/api-client-library/python/guide/aaa_oauth
# pip install google-api-python-client (1.6.4)

import httplib2
from apiclient import discovery
from oauth2client import client, tools
from oauth2client.file import Storage
from oauth2client.tools import argparser

HOME_DIR = os.path.expanduser('~')
GOOGLE_SECRETS = os.path.join(HOME_DIR, '.cache/secrets/client_secrets.json')


def get_credentials(*, scopes):
    """Gets valid user credentials from storage.

    If nothing has been stored, or if the stored credentials are invalid,
    the OAuth2 flow is completed to obtain the new credentials.

    Returns:
        Credentials, the obtained credential.
    """
    credential_dir = os.path.join(HOME_DIR, '.cache', __package__, 'credentials')
    if not os.path.exists(credential_dir):
        os.makedirs(credential_dir)
    credential_path = os.path.join(credential_dir, 'googleapis.json')

    store = Storage(credential_path)
    credentials = store.get()

    # see https://developers.google.com/api-client-library/python/auth/web-app
    # kw: "incremental scopes"
    if not credentials or credentials.invalid or not credentials.has_scopes(scopes):
        flow = client.flow_from_clientsecrets(GOOGLE_SECRETS, scopes)
        flow.user_agent = 'Bonobo ETL (https://www.bonobo-project.org/)'
        flags = argparser.parse_args(['--noauth_local_webserver'])
        credentials = tools.run_flow(flow, store, flags)
        print('Storing credentials to ' + credential_path)
    return credentials


def get_google_spreadsheets_api_client(scopes=('https://www.googleapis.com/auth/spreadsheets', )):
    credentials = get_credentials(scopes=scopes)
    http = credentials.authorize(httplib2.Http())
    discoveryUrl = 'https://sheets.googleapis.com/$discovery/rest?version=v4'
    return discovery.build('sheets', 'v4', http=http, discoveryServiceUrl=discoveryUrl, cache_discovery=False)


def get_google_people_api_client(scopes=('https://www.googleapis.com/auth/contacts', )):
    credentials = get_credentials(scopes=scopes)
    http = credentials.authorize(httplib2.Http())
    discoveryUrl = 'https://people.googleapis.com/$discovery/rest?version=v1'
    return discovery.build('people', 'v1', http=http, discoveryServiceUrl=discoveryUrl, cache_discovery=False)



'''
This example shows how a different file system service can be injected
into a transformation (as compressing pickled objects often makes sense
anyways).  The pickle itself contains a list of lists as follows:

```
[
    ['category', 'sms'],
    ['ham', 'Go until jurong point, crazy..'],
    ['ham', 'Ok lar... Joking wif u oni...'],
    ['spam', 'Free entry in 2 a wkly comp to win...'],
    ['ham', 'U dun say so early hor... U c already then say...'],
    ['ham', 'Nah I don't think he goes to usf, he lives around here though'],
    ['spam', 'FreeMsg Hey there darling it's been 3 week's now...'],
    ...
]
```

where the first column categorizes and sms as "ham" or "spam".  The second
column contains the sms itself.

Data set taken from:
https://www.kaggle.com/uciml/sms-spam-collection-dataset/downloads/sms-spam-collection-dataset.zip

The transformation (1) reads the pickled data, (2) marks and shortens
messages categorized as spam, and (3) prints the output.

'''

from fs.tarfs import TarFS

import bonobo
from bonobo import examples


def cleanse_sms(category, sms):
    if category == 'spam':
        sms_clean = '**MARKED AS SPAM** ' + sms[0:50] + (
            '...' if len(sms) > 50 else ''
        )
    elif category == 'ham':
        sms_clean = sms
    else:
        raise ValueError('Unknown category {!r}.'.format(category))

    return category, sms, sms_clean


def get_graph(*, _limit=(), _print=()):
    graph = bonobo.Graph()

    graph.add_chain(
        # spam.pkl is within the gzipped tarball
        bonobo.PickleReader('spam.pkl'),
        *_limit,
        cleanse_sms,
        *_print,
    )

    return graph


def get_services():
    from ._services import get_services
    return {
        **get_services(),
        'fs': TarFS(bonobo.get_examples_path('datasets/spam.tgz'))
    }


if __name__ == '__main__':
    parser = examples.get_argument_parser()
    with bonobo.parse_args(parser) as options:
        bonobo.run(
            get_graph(**examples.get_graph_options(options)),
            services=get_services()
        )



from bonobo.errors import AbstractError
from bonobo.util import isoption, iscontextprocessor, sortedlist, get_name

__all__ = [
    'Configurable',
]

get_creation_counter = lambda v: v._creation_counter


class ConfigurableMeta(type):
    """
    Metaclass for Configurables that will add options to a special __options__ dict.
    """

    def __init__(cls, what, bases=None, dict=None):
        super().__init__(what, bases, dict)

        cls.__processors = sortedlist()
        cls.__processors_cache = None
        cls.__methods = sortedlist()
        cls.__options = sortedlist()
        cls.__names = set()

        # cls.__kwoptions = []

        for typ in cls.__mro__:
            for name, value in filter(lambda x: isoption(x[1]), typ.__dict__.items()):
                if iscontextprocessor(value):
                    cls.__processors.insort((value._creation_counter, value))
                    continue

                if not value.name:
                    value.name = name

                if not name in cls.__names:
                    cls.__names.add(name)
                    cls.__options.insort((not value.positional, value._creation_counter, name, value))

        # Docstring formating
        _options_doc = []
        for _positional, _counter, _name, _value in cls.__options:
            _param = _name
            if _value.type:
                _param = get_name(_value.type) + ' ' + _param

            prefix = ':param {}: '.format(_param)
            for lineno, line in enumerate((_value.__doc__ or '').split('\n')):
                _options_doc.append((' ' * len(prefix) if lineno else prefix) + line)
        cls.__doc__ = '\n\n'.join(
            map(
                str.strip,
                filter(None, (
                    cls.__doc__,
                    '\n'.join(_options_doc)
                ))
            )
        )

    @property
    def __options__(cls):
        return ((name, option) for _, _, name, option in cls.__options)

    @property
    def __options_dict__(cls):
        return dict(cls.__options__)

    @property
    def __processors__(cls):
        if cls.__processors_cache is None:
            cls.__processors_cache = [processor for _, processor in cls.__processors]
        return cls.__processors_cache

    def __repr__(self):
        return ' '.join((
            '<Configurable',
            super(ConfigurableMeta, self).__repr__().split(' ', 1)[1],
        ))


try:
    import _functools
except:
    import functools

    PartiallyConfigured = functools.partial
else:

    class PartiallyConfigured(_functools.partial):
        @property  # TODO XXX cache this
        def _options_values(self):
            """ Simulate option values for partially configured objects. """
            try:
                return self.__options_values
            except AttributeError:
                self.__options_values = {**self.keywords}

                position = 0

                for name, option in self.func.__options__:
                    if not option.positional:
                        break  # no positional left
                    if name in self.keywords:
                        continue  # already fulfilled

                    self.__options_values[name] = self.args[position] if len(self.args) >= position + 1 else None
                    position += 1

                return self.__options_values

        def __getattr__(self, item):
            _dict = self.func.__options_dict__
            if item in _dict:
                return _dict[item].__get__(self, self.func)
            return getattr(self.func, item)


class Configurable(metaclass=ConfigurableMeta):
    """
    Generic class for configurable objects. Configurable objects have a dictionary of "options" descriptors that defines
    the configuration schema of the type.

    """

    def __new__(cls, *args, _final=False, **kwargs):
        """
        Custom instance builder. If not all options are fulfilled, will return a :class:`PartiallyConfigured` instance
        which is just a :class:`functools.partial` object that behaves like a :class:`Configurable` instance.

        The special `_final` argument can be used to force final instance to be created, or an error raised if options
        are missing.

        :param args:
        :param _final: bool
        :param kwargs:
        :return: Configurable or PartiallyConfigured
        """
        options = tuple(cls.__options__)
        # compute missing options, given the kwargs.
        missing = set()
        for name, option in options:
            if option.required and not option.name in kwargs:
                missing.add(name)

        # transform positional arguments in keyword arguments if possible.
        position = 0
        for name, option in options:
            if not option.positional:
                break  # option orders make all positional options first, job done.

            if not isoption(getattr(cls, name)):
                missing.remove(name)
                continue

            if len(args) <= position:
                break  # no more positional arguments given.

            position += 1
            if name in missing:
                missing.remove(name)

        # complain if there is more options than possible.
        extraneous = set(kwargs.keys()) - (set(next(zip(*options))) if len(options) else set())
        if len(extraneous):
            raise TypeError(
                '{}() got {} unexpected option{}: {}.'.format(
                    cls.__name__, len(extraneous), 's'
                    if len(extraneous) > 1 else '', ', '.join(map(repr, sorted(extraneous)))
                )
            )

        # missing options? we'll return a partial instance to finish the work later, unless we're required to be
        # "final".
        if len(missing):
            if _final:
                raise TypeError(
                    '{}() missing {} required option{}: {}.'.format(
                        cls.__name__, len(missing), 's'
                        if len(missing) > 1 else '', ', '.join(map(repr, sorted(missing)))
                    )
                )
            return PartiallyConfigured(cls, *args, **kwargs)

        return super(Configurable, cls).__new__(cls)

    def __init__(self, *args, **kwargs):
        # initialize option's value dictionary, used by descriptor implementation (see Option).
        self._options_values = {**kwargs}

        # set option values.
        for name, value in kwargs.items():
            setattr(self, name, value)

        position = 0
        for name, option in self.__options__:
            if not option.positional:
                break  # option orders make all positional options first

            # value was overriden? Skip.
            maybe_value = getattr(type(self), name)
            if not isoption(maybe_value):
                continue

            if len(args) <= position:
                break

            if name in self._options_values:
                raise ValueError('Already got a value for option {}'.format(name))

            setattr(self, name, args[position])
            position += 1

    def __call__(self, *args, **kwargs):
        raise AbstractError(self.__call__)

    @property
    def __options__(self):
        return type(self).__options__

    @property
    def __processors__(self):
        return type(self).__processors__



from bonobo.config import Configurable, ContextProcessor, Option, Service


class FileHandler(Configurable):
    """Abstract component factory for file-related components.

    Args:
        fs (str): service name to use for filesystem.
        path (str): which path to use within the provided filesystem.
        eol (str): which character to use to separate lines.
        mode (str): which mode to use when opening the file.
        encoding (str): which encoding to use when opening the file.
    """

    path = Option(str, required=True, positional=True, __doc__='''
        Path to use within the provided filesystem.
    ''')  # type: str
    eol = Option(str, default='\n', __doc__='''
        Character to use as line separator.
    ''')  # type: str
    mode = Option(str, __doc__='''
        What mode to use for open() call.
    ''')  # type: str
    encoding = Option(str, default='utf-8', __doc__='''
        Encoding.
    ''')  # type: str
    fs = Service('fs', __doc__='''
        The filesystem instance to use.
    ''')  # type: str

    @ContextProcessor
    def file(self, context, *, fs):
        with self.open(fs) as file:
            yield file

    def open(self, fs):
        return fs.open(self.path, self.mode, encoding=self.encoding)


class Reader:
    pass


class Writer:
    pass



from collections import Iterable
from contextlib import contextmanager
from functools import partial
from inspect import signature

from bonobo.config import Option
from bonobo.errors import UnrecoverableTypeError
from bonobo.util import deprecated_alias, ensure_tuple

_raw = object()
_args = object()
_none = object()

INPUT_FORMATS = {_raw, _args, _none}


class ContextProcessor(Option):
    """
    A ContextProcessor is a kind of transformation decorator that can setup and teardown a transformation and runtime
    related dependencies, at the execution level.

    It works like a yielding context manager, and is the recommended way to setup and teardown objects you'll need
    in the context of one execution. It's the way to overcome the stateless nature of transformations.

    The yielded values will be passed as positional arguments to the next context processors (order does matter), and
    finally to the __call__ method of the transformation.

    Warning: this may change for a similar but simpler implementation, don't rely too much on it (yet).

    Example:

        >>> from bonobo.config import Configurable
        >>> from bonobo.util import ValueHolder

        >>> class Counter(Configurable):
        ...     @ContextProcessor
        ...     def counter(self, context):
        ...         yield ValueHolder(0)
        ...
        ...     def __call__(self, counter, *args, **kwargs):
        ...         counter += 1
        ...         yield counter.get()

    """

    @property
    def __name__(self):
        return self.func.__name__

    def __init__(self, func):
        self.func = func
        super(ContextProcessor, self).__init__(required=False, default=self.__name__)
        self.name = self.__name__

    def __repr__(self):
        return repr(self.func).replace('<function', '<{}'.format(type(self).__name__))

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class bound(partial):
    @property
    def kwargs(self):
        return self.keywords


class ContextCurrifier:
    """
    This is a helper to resolve processors.
    """

    def __init__(self, wrapped, *args, **kwargs):
        self.wrapped = wrapped
        self.args = args
        self.kwargs = kwargs
        self.format = getattr(wrapped, '__input_format__', _args)
        self._stack, self._stack_values = None, None

    def __iter__(self):
        yield from self.wrapped

    def _bind(self, _input):
        try:
            bind = signature(self.wrapped).bind
        except ValueError:
            bind = partial(bound, self.wrapped)
        if self.format is _args:
            return bind(*self.args, *_input, **self.kwargs)
        if self.format is _raw:
            return bind(*self.args, _input, **self.kwargs)
        if self.format is _none:
            return bind(*self.args, **self.kwargs)
        raise NotImplementedError('Invalid format {!r}.'.format(self.format))

    def __call__(self, _input):
        if not callable(self.wrapped):
            if isinstance(self.wrapped, Iterable):
                return self.__iter__()
            raise UnrecoverableTypeError('Uncallable node {}'.format(self.wrapped))
        try:
            bound = self._bind(_input)
        except TypeError as exc:
            raise UnrecoverableTypeError((
                'Input of {wrapped!r} does not bind to the node signature.\n'
                'Args: {args}\n'
                'Input: {input}\n'
                'Kwargs: {kwargs}\n'
                'Signature: {sig}'
            ).format(
                wrapped=self.wrapped, args=self.args, input=_input, kwargs=self.kwargs, sig=signature(self.wrapped)
            )) from exc
        return self.wrapped(*bound.args, **bound.kwargs)

    def setup(self, *context):
        if self._stack is not None:
            raise RuntimeError('Cannot setup context currification twice.')

        self._stack, self._stack_values = list(), list()
        for processor in resolve_processors(self.wrapped):
            _processed = processor(self.wrapped, *context, *self.args, **self.kwargs)
            _append_to_context = next(_processed)
            self._stack_values.append(_append_to_context)
            if _append_to_context is not None:
                self.args += ensure_tuple(_append_to_context)
            self._stack.append(_processed)

    def teardown(self):
        while self._stack:
            processor = self._stack.pop()
            try:
                # todo yield from ? how to ?
                processor.send(self._stack_values.pop())
            except StopIteration as exc:
                # This is normal, and wanted.
                pass
            else:
                # No error ? We should have had StopIteration ...
                raise RuntimeError('Context processors should not yield more than once.')
        self._stack, self._stack_values = None, None

    @contextmanager
    def as_contextmanager(self, *context):
        """
        Convenience method to use it as a contextmanager, mostly for test purposes.

        Example:

            >>> with ContextCurrifier(node).as_contextmanager(context) as stack:
            ...     stack()

        :param context:
        :return:
        """
        self.setup(*context)
        yield self
        self.teardown()


def resolve_processors(mixed):
    try:
        yield from mixed.__processors__
    except AttributeError:
        yield from ()


get_context_processors = deprecated_alias('get_context_processors', resolve_processors)


def use_context(f):
    def context(self, context, *args, **kwargs):
        yield context

    return use_context_processor(context)(f)


def use_context_processor(context_processor):
    def using_context_processor(cls_or_func):
        nonlocal context_processor

        try:
            cls_or_func.__processors__
        except AttributeError:
            cls_or_func.__processors__ = []

        cls_or_func.__processors__.append(ContextProcessor(context_processor))
        return cls_or_func

    return using_context_processor


def _use_input_format(input_format):
    if input_format not in INPUT_FORMATS:
        raise ValueError(
            'Invalid input format {!r}. Choices: {}'.format(input_format, ', '.join(sorted(INPUT_FORMATS)))
        )

    def _set_input_format(f):
        setattr(f, '__input_format__', input_format)
        return f

    return _set_input_format


use_no_input = _use_input_format(_none)
use_raw_input = _use_input_format(_raw)



"""
Example on how to use symple python strings to communicate between transformations.

.. graphviz::

    digraph {
        rankdir = LR;
        stylesheet = "../_static/graphs.css";

        BEGIN [shape="point"];
        BEGIN -> "extract()" -> "transform(s: str)" -> "load(s: str)";
    }

"""
from random import randint

import bonobo


def extract():
    yield 'foo'
    yield 'bar'
    yield 'baz'


def transform(s):
    return '{} ({})'.format(s.title(), randint(10, 99))


def load(s):
    print(s)


def get_graph():
    return bonobo.Graph(extract, transform, load)


if __name__ == '__main__':
    parser = bonobo.get_argument_parser()
    with bonobo.parse_args(parser):
        bonobo.run(get_graph())



import pkg_resources
from packaging.utils import canonicalize_name

bonobo_packages = {}
for p in pkg_resources.working_set:
    name = canonicalize_name(p.project_name)
    if name.startswith('bonobo'):
        bonobo_packages[name] = p



from bonobo.execution.contexts.graph import GraphExecutionContext


class Strategy:
    """
    Base class for execution strategies.

    """
    GraphExecutionContextType = GraphExecutionContext

    def __init__(self, GraphExecutionContextType=None):
        self.GraphExecutionContextType = GraphExecutionContextType or self.GraphExecutionContextType

    def create_graph_execution_context(self, graph, *args, GraphExecutionContextType=None, **kwargs):
        return (GraphExecutionContextType or self.GraphExecutionContextType)(graph, *args, **kwargs)

    def execute(self, graph, *args, **kwargs):
        raise NotImplementedError



import html
import json
from collections import namedtuple
from copy import copy

from bonobo.constants import BEGIN
from bonobo.util import get_name
from graphviz import ExecutableNotFound
from graphviz.dot import Digraph

GraphRange = namedtuple('GraphRange', ['graph', 'input', 'output'])


class Graph:
    """
    Represents a directed graph of nodes.
    """
    name = ''

    def __init__(self, *chain):
        self.edges = {BEGIN: set()}
        self.named = {}
        self.nodes = []
        self.add_chain(*chain)

    def __iter__(self):
        yield from self.nodes

    def __len__(self):
        """ Node count.
        """
        return len(self.nodes)

    def __getitem__(self, key):
        return self.nodes[key]

    def outputs_of(self, idx, create=False):
        """ Get a set of the outputs for a given node index.
        """
        if create and not idx in self.edges:
            self.edges[idx] = set()
        return self.edges[idx]

    def add_node(self, c):
        """ Add a node without connections in this graph and returns its index.
        """
        idx = len(self.nodes)
        self.edges[idx] = set()
        self.nodes.append(c)
        return idx

    def add_chain(self, *nodes, _input=BEGIN, _output=None, _name=None):
        """ Add a chain in this graph.
        """
        if len(nodes):
            _input = self._resolve_index(_input)
            _output = self._resolve_index(_output)
            _first = None
            _last = None

            for i, node in enumerate(nodes):
                _last = self.add_node(node)
                if not i and _name:
                    if _name in self.named:
                        raise KeyError('Duplicate name {!r} in graph.'.format(_name))
                    self.named[_name] = _last
                if _first is None:
                    _first = _last
                self.outputs_of(_input, create=True).add(_last)
                _input = _last

            if _output is not None:
                self.outputs_of(_input, create=True).add(_output)

            if hasattr(self, '_topologcally_sorted_indexes_cache'):
                del self._topologcally_sorted_indexes_cache

            return GraphRange(self, _first, _last)
        return GraphRange(self, None, None)

    def copy(self):
        g = Graph()

        g.edges = copy(self.edges)
        g.named = copy(self.named)
        g.nodes = copy(self.nodes)

        return g

    @property
    def topologically_sorted_indexes(self):
        """Iterate in topological order, based on networkx's topological_sort() function.
        """
        try:
            return self._topologcally_sorted_indexes_cache
        except AttributeError:
            seen = set()
            order = []
            explored = set()

            for i in self.edges:
                if i in explored:
                    continue
                fringe = [i]
                while fringe:
                    w = fringe[-1]  # depth first search
                    if w in explored:  # already looked down this branch
                        fringe.pop()
                        continue
                    seen.add(w)  # mark as seen
                    # Check successors for cycles and for new nodes
                    new_nodes = []
                    for n in self.outputs_of(w):
                        if n not in explored:
                            if n in seen:  # CYCLE !!
                                raise RuntimeError("Graph contains a cycle.")
                            new_nodes.append(n)
                    if new_nodes:  # Add new_nodes to fringe
                        fringe.extend(new_nodes)
                    else:  # No new nodes so w is fully explored
                        explored.add(w)
                        order.append(w)
                        fringe.pop()  # done considering this node
            self._topologcally_sorted_indexes_cache = tuple(filter(lambda i: type(i) is int, reversed(order)))
            return self._topologcally_sorted_indexes_cache

    @property
    def graphviz(self):
        try:
            return self._graphviz
        except AttributeError:
            g = Digraph()
            g.attr(rankdir='LR')
            g.node('BEGIN', shape='point')
            for i in self.outputs_of(BEGIN):
                g.edge('BEGIN', str(i))
            for ix in self.topologically_sorted_indexes:
                g.node(str(ix), label=get_name(self[ix]))
                for iy in self.outputs_of(ix):
                    g.edge(str(ix), str(iy))
            self._graphviz = g
            return self._graphviz

    def _repr_dot_(self):
        return str(self.graphviz)

    def _repr_html_(self):
        try:
            return '<div>{}</div><pre>{}</pre>'.format(self.graphviz._repr_svg_(), html.escape(repr(self)))
        except (ExecutableNotFound, FileNotFoundError) as exc:
            return '<strong>{}</strong>: {}'.format(type(exc).__name__, str(exc))

    def _resolve_index(self, mixed):
        """ Find the index based on various strategies for a node, probably an input or output of chain. Supported inputs are indexes, node values or names.
        """
        if mixed is None:
            return None

        if type(mixed) is int or mixed in self.edges:
            return mixed

        if isinstance(mixed, str) and mixed in self.named:
            return self.named[mixed]

        if mixed in self.nodes:
            return self.nodes.index(mixed)

        raise ValueError('Cannot find node matching {!r}.'.format(mixed))


def _get_graphviz_node_id(graph, i):
    escaped_index = str(i)
    escaped_name = json.dumps(get_name(graph[i]))
    return '{{{} [label={}]}}'.format(escaped_index, escaped_name)



from bonobo.commands import BaseCommand

all_examples = (
    'clock',
    'datasets',
    'environ',
    'files.csv_handlers',
    'files.json_handlers',
    'files.pickle_handlers',
    'files.text_handlers',
    'types',
)


class ExamplesCommand(BaseCommand):
    def handle(self):
        print('You can run the following examples:')
        print()
        for example in all_examples:
            print('  $ python -m bonobo.examples.{}'.format(example))
        print()

    def add_arguments(self, parser):
        pass



import bisect
import functools


class sortedlist(list):
    def insort(self, x):
        bisect.insort(self, x)


def ensure_tuple(tuple_or_mixed, *, cls=tuple):
    """
    If it's not a tuple, let's make a tuple of one item.
    Otherwise, not changed.

    :param tuple_or_mixed:
    :return: tuple

    """

    if isinstance(tuple_or_mixed, cls):
        return tuple_or_mixed

    if tuple_or_mixed is None:
        return tuple.__new__(cls, ())

    if isinstance(tuple_or_mixed, tuple):
        return tuple.__new__(cls, tuple_or_mixed)

    return tuple.__new__(cls, (tuple_or_mixed,))


def cast(type_):
    def _wrap_cast(f):
        @functools.wraps(f)
        def _wrapped_cast(*args, **kwargs):
            nonlocal f, type_
            return type_(f(*args, **kwargs))

        return _wrapped_cast

    return _wrap_cast


tuplize = cast(tuple)
tuplize.__doc__ = """
Decorates a generator and make it a tuple-returning function. As a side effect, it can also decorate any
iterator-returning function to force return value to be a tuple.

>>> tuplized_lambda = tuplize(lambda: [1, 2, 3])
>>> tuplized_lambda()
(1, 2, 3)

>>> @tuplize
... def my_generator():
...     yield 1
...     yield 2
...     yield 3
...
>>> my_generator()
(1, 2, 3)
"""



import os

import bonobo


def get_minor_version():
    return '.'.join(bonobo.__version__.split('.')[:2])


def get_datasets_dir(*dirs):
    home_dir = os.path.expanduser('~')
    target_dir = os.path.join(
        home_dir, '.cache/bonobo', get_minor_version(), *dirs
    )
    os.makedirs(target_dir, exist_ok=True)
    return target_dir


def get_services():
    return {'fs': bonobo.open_fs(get_datasets_dir('datasets'))}



import functools
import html
import itertools
import pprint

from bonobo import settings
from bonobo.config import Configurable, Option, Method, use_raw_input, use_context, use_no_input
from bonobo.config.functools import transformation_factory
from bonobo.config.processors import ContextProcessor, use_context_processor
from bonobo.constants import NOT_MODIFIED
from bonobo.util.objects import ValueHolder
from bonobo.util.term import CLEAR_EOL
from mondrian import term

__all__ = [
    'FixedWindow',
    'Format',
    'Limit',
    'OrderFields',
    'PrettyPrinter',
    'Rename',
    'SetFields',
    'Tee',
    'UnpackItems',
    'count',
    'identity',
    'noop',
]


def identity(x):
    return x


class Limit(Configurable):
    """
    Creates a Limit() node, that will only let go through the first n rows (defined by the `limit` option), unmodified.

    .. attribute:: limit

        Number of rows to let go through.

    TODO: simplify into a closure building factory?

    """
    limit = Option(positional=True, default=10)

    @ContextProcessor
    def counter(self, context):
        yield ValueHolder(0)

    def __call__(self, counter, *args, **kwargs):
        counter += 1
        if counter <= self.limit:
            yield NOT_MODIFIED


def Tee(f):
    from bonobo.constants import NOT_MODIFIED

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        nonlocal f
        f(*args, **kwargs)
        return NOT_MODIFIED

    return wrapped


def _shorten(s, w):
    if w and len(s) > w:
        s = s[0:w - 3] + '...'
    return s


class PrettyPrinter(Configurable):
    max_width = Option(
        int,
        default=term.get_size()[0],
        required=False,
        __doc__='''
        If set, truncates the output values longer than this to this width.
    '''
    )

    filter = Method(
        default=
        (lambda self, index, key, value: (value is not None) and (not isinstance(key, str) or not key.startswith('_'))),
        __doc__='''
            A filter that determine what to print.
            
            Default is to ignore any key starting with an underscore and none values.
        '''
    )

    @ContextProcessor
    def context(self, context):
        context.setdefault('_jupyter_html', None)
        yield context
        if context._jupyter_html is not None:
            from IPython.display import display, HTML
            display(HTML('\n'.join(['<table>'] + context._jupyter_html + ['</table>'])))

    def __call__(self, context, *args, **kwargs):
        if not settings.QUIET:
            if term.isjupyter:
                self.print_jupyter(context, *args, **kwargs)
                return NOT_MODIFIED
            if term.istty:
                self.print_console(context, *args, **kwargs)
                return NOT_MODIFIED

        self.print_quiet(context, *args, **kwargs)
        return NOT_MODIFIED

    def print_quiet(self, context, *args, **kwargs):
        for index, (key, value) in enumerate(itertools.chain(enumerate(args), kwargs.items())):
            if self.filter(index, key, value):
                print(self.format_quiet(index, key, value, fields=context.get_input_fields()))

    def format_quiet(self, index, key, value, *, fields=None):
        # XXX should we implement argnames here ?
        return ' '.join(((' ' if index else '-'), str(key), ':', str(value).strip()))

    def print_console(self, context, *args, **kwargs):
        print('\u250c')
        for index, (key, value) in enumerate(itertools.chain(enumerate(args), kwargs.items())):
            if self.filter(index, key, value):
                print(self.format_console(index, key, value, fields=context.get_input_fields()))
        print('\u2514')

    def format_console(self, index, key, value, *, fields=None):
        fields = fields or []
        if not isinstance(key, str):
            if len(fields) > key and str(key) != str(fields[key]):
                key = '{}{}'.format(fields[key], term.lightblack('[{}]'.format(key)))
            else:
                key = str(index)

        prefix = '\u2502 {} = '.format(key)
        prefix_length = len(prefix)

        def indent(text, prefix):
            for i, line in enumerate(text.splitlines()):
                yield (prefix if i else '') + line + CLEAR_EOL + '\n'

        repr_of_value = ''.join(
            indent(pprint.pformat(value, width=self.max_width - prefix_length), '\u2502' + ' ' * (len(prefix) - 1))
        ).strip()
        return '{}{}{}'.format(prefix, repr_of_value.replace('\n', CLEAR_EOL + '\n'), CLEAR_EOL)

    def print_jupyter(self, context, *args):
        if not context._jupyter_html:
            context._jupyter_html = [
                '<thead><tr>',
                *map('<th>{}</th>'.format, map(html.escape, map(str,
                                                                context.get_input_fields() or range(len(args))))),
                '</tr></thead>',
            ]

        context._jupyter_html += [
            '<tr>',
            *map('<td>{}</td>'.format, map(html.escape, map(repr, args))),
            '</tr>',
        ]


@use_no_input
def noop(*args, **kwargs):
    return NOT_MODIFIED


class FixedWindow(Configurable):
    """
    Transformation factory to create fixed windows of inputs, as lists.

    For example, if the input is successively 1, 2, 3, 4, etc. and you pass it through a ``FixedWindow(2)``, you'll get
    lists of elements 2 by 2: [1, 2], [3, 4], ...

    """

    length = Option(int, positional=True)  # type: int

    @ContextProcessor
    def buffer(self, context):
        buffer = yield ValueHolder([])
        if len(buffer):
            last_value = buffer.get()
            last_value += [None] * (self.length - len(last_value))
            context.send(*last_value)

    @use_raw_input
    def __call__(self, buffer, bag):
        buffer.append(bag)
        if len(buffer) >= self.length:
            yield tuple(buffer.get())
            buffer.set([])


@transformation_factory
def OrderFields(fields):
    """
    Transformation factory to reorder fields in a data stream.

    :param fields:
    :return: callable
    """
    fields = list(fields)

    @use_context
    @use_raw_input
    def _OrderFields(context, row):
        nonlocal fields
        context.setdefault('remaining', None)
        if not context.output_type:
            context.remaining = list(sorted(set(context.get_input_fields()) - set(fields)))
            context.set_output_fields(fields + context.remaining)

        yield tuple(row.get(field) for field in context.get_output_fields())

    return _OrderFields


@transformation_factory
def SetFields(fields):
    """
    Transformation factory that sets the field names on first iteration, without touching the values.

    :param fields:
    :return: callable
    """

    @use_context
    @use_no_input
    def _SetFields(context):
        nonlocal fields
        if not context.output_type:
            context.set_output_fields(fields)
        return NOT_MODIFIED

    return _SetFields


@transformation_factory
def UnpackItems(*items, fields=None, defaults=None):
    """
    >>> UnpackItems(0)

    :param items:
    :param fields:
    :param defaults:
    :return: callable
    """
    defaults = defaults or {}

    @use_context
    @use_raw_input
    def _UnpackItems(context, bag):
        nonlocal fields, items, defaults

        if fields is None:
            fields = ()
            for item in items:
                fields += tuple(bag[item].keys())
            context.set_output_fields(fields)

        values = ()
        for item in items:
            values += tuple(bag[item].get(field, defaults.get(field)) for field in fields)

        return values

    return _UnpackItems


@transformation_factory
def Rename(**translations):
    # XXX todo handle duplicated

    fields = None
    translations = {v: k for k, v in translations.items()}

    @use_context
    @use_raw_input
    def _Rename(context, bag):
        nonlocal fields, translations

        if not fields:
            fields = tuple(translations.get(field, field) for field in context.get_input_fields())
            context.set_output_fields(fields)

        return NOT_MODIFIED

    return _Rename


@transformation_factory
def Format(**formats):
    fields, newfields = None, None

    @use_context
    @use_raw_input
    def _Format(context, bag):
        nonlocal fields, newfields, formats

        if not context.output_type:
            fields = context.input_type._fields
            newfields = tuple(field for field in formats if not field in fields)
            context.set_output_fields(fields + newfields)

        return tuple(
            formats[field].format(**bag._asdict()) if field in formats else bag.get(field)
            for field in fields + newfields
        )

    return _Format


def _count(self, context):
    counter = yield ValueHolder(0)
    context.send(counter.get())


@use_no_input
@use_context_processor(_count)
def count(counter):
    counter += 1



from bonobo.execution.strategies.executor import ProcessPoolExecutorStrategy, ThreadPoolExecutorStrategy
from bonobo.execution.strategies.naive import NaiveStrategy

__all__ = [
    'create_strategy',
]

STRATEGIES = {
    'naive': NaiveStrategy,
    'processpool': ProcessPoolExecutorStrategy,
    'threadpool': ThreadPoolExecutorStrategy,
}

DEFAULT_STRATEGY = 'threadpool'


def create_strategy(name=None):
    """
    Create a strategy, or just returns it if it's already one.
    
    :param name: 
    :return: Strategy
    """
    import logging
    from bonobo.execution.strategies.base import Strategy

    if isinstance(name, Strategy):
        return name

    if name is None:
        name = DEFAULT_STRATEGY

    logging.debug('Creating strategy {}...'.format(name))

    try:
        factory = STRATEGIES[name]
    except KeyError as exc:
        raise RuntimeError(
            'Invalid strategy {}. Available choices: {}.'.format(repr(name), ', '.join(sorted(STRATEGIES.keys())))
        ) from exc

    return factory()



from bonobo.execution.strategies import create_strategy
from bonobo.nodes import __all__ as _all_nodes
from bonobo.nodes import *
from bonobo.structs import Graph
from bonobo.util import get_name
from bonobo.util.api import ApiHelper
from bonobo.util.environ import parse_args, get_argument_parser

__all__ = []

api = ApiHelper(__all__)


@api.register_graph
def run(graph, *, plugins=None, services=None, strategy=None):
    """
    Main entry point of bonobo. It takes a graph and creates all the necessary plumbing around to execute it.

    The only necessary argument is a :class:`Graph` instance, containing the logic you actually want to execute.

    By default, this graph will be executed using the "threadpool" strategy: each graph node will be wrapped in a
    thread, and executed in a loop until there is no more input to this node.

    You can provide plugins factory objects in the plugins list, this function will add the necessary plugins for
    interactive console execution and jupyter notebook execution if it detects correctly that it runs in this context.

    You'll probably want to provide a services dictionary mapping service names to service instances.

    :param Graph graph: The :class:`Graph` to execute.
    :param str strategy: The :class:`bonobo.execution.strategies.base.Strategy` to use.
    :param list plugins: The list of plugins to enhance execution.
    :param dict services: The implementations of services this graph will use.
    :return bonobo.execution.graph.GraphExecutionContext:
    """

    plugins = plugins or []

    from bonobo import settings
    settings.check()

    if not settings.QUIET.get():  # pragma: no cover
        if _is_interactive_console():
            import mondrian
            mondrian.setup(excepthook=True)

            from bonobo.plugins.console import ConsoleOutputPlugin
            if ConsoleOutputPlugin not in plugins:
                plugins.append(ConsoleOutputPlugin)

        if _is_jupyter_notebook():
            try:
                from bonobo.contrib.jupyter import JupyterOutputPlugin
            except ImportError:
                import logging
                logging.warning(
                    'Failed to load jupyter widget. Easiest way is to install the optional "jupyter" '
                    'dependencies with «pip install bonobo[jupyter]», but you can also install a specific '
                    'version by yourself.'
                )
            else:
                if JupyterOutputPlugin not in plugins:
                    plugins.append(JupyterOutputPlugin)

    import logging
    logging.getLogger().setLevel(settings.LOGGING_LEVEL.get())
    strategy = create_strategy(strategy)
    return strategy.execute(graph, plugins=plugins, services=services)


def _inspect_as_graph(graph):
    return graph._repr_dot_()


_inspect_formats = {'graph': _inspect_as_graph}


@api.register_graph
def inspect(graph, *, plugins=None, services=None, strategy=None, format):
    if not format in _inspect_formats:
        raise NotImplementedError(
            'Output format {} not implemented. Choices are: {}.'.format(
                format, ', '.join(sorted(_inspect_formats.keys()))
            )
        )
    print(_inspect_formats[format](graph))


# data structures
api.register_group(Graph)

# execution strategies
api.register_group(create_strategy)


# Shortcut to filesystem2's open_fs, that we make available there for convenience.
@api.register
def open_fs(fs_url=None, *args, **kwargs):
    """
    Wraps :func:`fs.open_fs` function with a few candies.

    :param str fs_url: A filesystem URL
    :param parse_result: A parsed filesystem URL.
    :type parse_result: :class:`ParseResult`
    :param bool writeable: True if the filesystem must be writeable.
    :param bool create: True if the filesystem should be created if it does not exist.
    :param str cwd: The current working directory (generally only relevant for OS filesystems).
    :param str default_protocol: The protocol to use if one is not supplied in the FS URL (defaults to ``"osfs"``).
    :returns: :class:`~fs.base.FS` object
    """
    from fs import open_fs as _open_fs
    from os.path import expanduser
    from os import getcwd

    if fs_url is None:
        fs_url = getcwd()

    return _open_fs(expanduser(str(fs_url)), *args, **kwargs)


# standard transformations
api.register_group(
    CsvReader,
    CsvWriter,
    FileReader,
    FileWriter,
    Filter,
    FixedWindow,
    Format,
    JsonReader,
    JsonWriter,
    LdjsonReader,
    LdjsonWriter,
    Limit,
    OrderFields,
    PickleReader,
    PickleWriter,
    PrettyPrinter,
    RateLimited,
    Rename,
    SetFields,
    Tee,
    UnpackItems,
    count,
    identity,
    noop,
    check=_all_nodes,
)


def _is_interactive_console():
    import sys
    return sys.stdout.isatty()


def _is_jupyter_notebook():
    try:
        return get_ipython().__class__.__name__ == 'ZMQInteractiveShell'
    except NameError:
        return False


@api.register
def get_examples_path(*pathsegments):
    import os
    import pathlib
    return str(pathlib.Path(os.path.dirname(__file__), 'examples', *pathsegments))


@api.register
def open_examples_fs(*pathsegments):
    return open_fs(get_examples_path(*pathsegments))


api.register_group(get_argument_parser, parse_args)



import argparse
import logging
import runpy
import sys
from contextlib import contextmanager

import bonobo.util.environ
from bonobo.util import get_name
from bonobo.util.environ import get_argument_parser, parse_args


class BaseCommand:
    """
    Base class for CLI commands.

    """

    @property
    def logger(self):
        try:
            return self._logger
        except AttributeError:
            self._logger = logging.getLogger(get_name(self))
            return self._logger

    def add_arguments(self, parser):
        """
        Entry point for subclassed commands to add custom arguments.
        """
        pass

    def handle(self, *args, **options):
        """
        The actual logic of the command. Subclasses must implement this method.
        """
        raise NotImplementedError('Subclasses of BaseCommand must provide a handle() method')


class BaseGraphCommand(BaseCommand):
    """
    Base class for CLI commands that depends on a graph definition, either from a file or from a module.

    """
    required = True
    handler = None

    def add_arguments(self, parser):
        # target arguments (cannot provide both).
        source_group = parser.add_mutually_exclusive_group(required=self.required)
        source_group.add_argument('file', nargs='?', type=str)
        source_group.add_argument('-m', dest='mod', type=str)

        # add arguments to enforce system environment.
        parser = get_argument_parser(parser)

        return parser

    def parse_options(self, **options):
        return options

    def handle(self, file, mod, **options):
        options = self.parse_options(**options)
        with self.read(file, mod, **options) as (graph, graph_execution_options, options):
            return self.do_handle(graph, **graph_execution_options, **options)

    def do_handle(self, graph, **options):
        if not self.handler:
            raise RuntimeError('{} has no handler defined.'.format(get_name(self)))
        return self.handler(graph, **options)

    @contextmanager
    def read(self, file, mod, **options):
        _graph, _graph_execution_options = None, None

        def _record(graph, **graph_execution_options):
            nonlocal _graph, _graph_execution_options
            _graph, _graph_execution_options = graph, graph_execution_options

        with _override_runner(_record), parse_args(options) as options:
            _argv = sys.argv
            try:
                if file:
                    sys.argv = [file]
                    self._run_path(file)
                elif mod:
                    sys.argv = [mod]
                    self._run_module(mod)
                else:
                    raise RuntimeError('No target provided.')
            finally:
                sys.argv = _argv

            if _graph is None:
                raise RuntimeError('Could not find graph.')

            yield _graph, _graph_execution_options, options

    def _run_path(self, file):
        return runpy.run_path(file, run_name='__main__')

    def _run_module(self, mod):
        return runpy.run_module(mod, run_name='__main__')


@contextmanager
def _override_runner(runner):
    """
    Context manager that monkey patches `bonobo.run` function with our current command logic.

    :param runner: the callable that will handle the `run()` logic.
    """
    import bonobo

    _get_argument_parser = bonobo.util.environ.get_argument_parser
    _run = bonobo.run
    try:
        # Original get_argument_parser would create or update an argument parser with environment options, but here we
        # already had them parsed so let's patch with something that creates an empty one instead.
        def get_argument_parser(parser=None):
            return parser or argparse.ArgumentParser()

        bonobo.util.environ.get_argument_parser = get_argument_parser
        bonobo.run = runner

        yield runner
    finally:
        # Restore our saved values.
        bonobo.util.environ.get_argument_parser = _get_argument_parser
        bonobo.run = _run



from bonobo.util.collections import cast, ensure_tuple, sortedlist, tuplize
from bonobo.util.compat import deprecated, deprecated_alias
from bonobo.util.inspect import (
    inspect_node,
    isconfigurable,
    isconfigurabletype,
    iscontextprocessor,
    isdict,
    ismethod,
    isoption,
    istuple,
    istype,
)
from bonobo.util.objects import (get_name, get_attribute_or_create, ValueHolder)

# Bonobo's util API
__all__ = [
    'ValueHolder',
    'cast',
    'deprecated',
    'deprecated_alias',
    'ensure_tuple',
    'get_attribute_or_create',
    'get_name',
    'inspect_node',
    'isconfigurable',
    'isconfigurabletype',
    'iscontextprocessor',
    'isdict',
    'ismethod',
    'isoption',
    'istype',
    'sortedlist',
    'tuplize',
]



import io
import re

import requests

import bonobo
from bonobo.commands import BaseCommand

EXAMPLES_BASE_URL = 'https://raw.githubusercontent.com/python-bonobo/bonobo/master/bonobo/examples/'
"""The URL to our git repository, in raw mode."""


class DownloadCommand(BaseCommand):
    def handle(self, *, path, **options):
        if not path.startswith('examples'):
            raise ValueError('Download command currently supports examples only')
        examples_path = re.sub('^examples/', '', path)
        output_path = bonobo.get_examples_path(examples_path)
        with _open_url(EXAMPLES_BASE_URL + examples_path) as response, open(output_path, 'wb') as fout:
            for chunk in response.iter_content(io.DEFAULT_BUFFER_SIZE):
                fout.write(chunk)
        self.logger.info('Download saved to {}'.format(output_path))

    def add_arguments(self, parser):
        parser.add_argument('path', help='The relative path of the thing to download.')


def _open_url(url):
    """Open a HTTP connection to the URL and return a file-like object."""
    response = requests.get(url, stream=True)
    if response.status_code != 200:
        raise IOError('Unable to download {}, HTTP {}'.format(url, response.status_code))
    return response



import functools
import logging
import sys
from concurrent.futures import Executor, ProcessPoolExecutor, ThreadPoolExecutor

from bonobo.constants import BEGIN, END
from bonobo.execution.strategies.base import Strategy

logger = logging.getLogger(__name__)


class ExecutorStrategy(Strategy):
    """
    Strategy based on a concurrent.futures.Executor subclass (or similar interface).

    """

    executor_factory = Executor

    def create_executor(self):
        return self.executor_factory()

    def execute(self, graph, **kwargs):
        context = self.create_graph_execution_context(graph, **kwargs)
        context.write(BEGIN, (), END)

        futures = []

        with self.create_executor() as executor:
            try:
                context.start(self.get_starter(executor, futures))
            except:
                logger.critical('Exception caught while starting execution context.', exc_info=sys.exc_info())

            while context.alive:
                try:
                    context.tick()
                except KeyboardInterrupt:
                    logging.getLogger(__name__).warning(
                        'KeyboardInterrupt received. Trying to terminate the nodes gracefully.'
                    )
                    context.kill()
                    break

            context.stop()

        return context

    def get_starter(self, executor, futures):
        def starter(node):
            @functools.wraps(node)
            def _runner():
                try:
                    with node:
                        node.loop()
                except:
                    logging.getLogger(__name__).critical(
                        'Critical error in threadpool node starter.', exc_info=sys.exc_info()
                    )

            try:
                futures.append(executor.submit(_runner))
            except:
                logging.getLogger(__name__).critical('futures.append', exc_info=sys.exc_info())

        return starter


class ThreadPoolExecutorStrategy(ExecutorStrategy):
    executor_factory = ThreadPoolExecutor


class ProcessPoolExecutorStrategy(ExecutorStrategy):
    executor_factory = ProcessPoolExecutor



from bonobo.constants import NOT_MODIFIED

from bonobo.config import Configurable, Method


class Filter(Configurable):
    """Filter out hashes from the stream depending on the :attr:`filter` callable return value, when called with the
    current hash as parameter.

    Can be used as a decorator on a filter callable.

    .. attribute:: filter

        A callable used to filter lines.
        
        If the callable returns a true-ish value, the input will be passed unmodified to the next items.
        
        Otherwise, it'll be burnt.

    """

    filter = Method()

    def __call__(self, *args, **kwargs):
        if self.filter(*args, **kwargs):
            return NOT_MODIFIED



from bonobo.commands import BaseCommand


def get_versions(*, all=False, quiet=None):
    import bonobo
    from bonobo.util.pkgs import bonobo_packages

    yield _format_version(bonobo, quiet=quiet)

    if all:
        for name in sorted(bonobo_packages):
            if name != 'bonobo':
                try:
                    mod = __import__(name.replace('-', '_'))
                    try:
                        yield _format_version(mod, name=name, quiet=quiet)
                    except Exception as exc:
                        yield '{} ({})'.format(name, exc)
                except ImportError as exc:
                    yield '{} is not importable ({}).'.format(name, exc)


class VersionCommand(BaseCommand):
    def handle(self, *, all=False, quiet=False):
        for line in get_versions(all=all, quiet=quiet):
            print(line)

    def add_arguments(self, parser):
        parser.add_argument('--all', '-a', action='store_true')
        parser.add_argument('--quiet', '-q', action='count')


def _format_version(mod, *, name=None, quiet=False):
    from bonobo.util.pkgs import bonobo_packages
    args = {
        'name': name or mod.__name__,
        'version': mod.__version__,
        'location': bonobo_packages[name or mod.__name__].location
    }

    if not quiet:
        return '{name} v.{version} (in {location})'.format(**args)
    if quiet < 2:
        return '{name} {version}'.format(**args)
    if quiet < 3:
        return '{version}'.format(**args)

    raise RuntimeError('Hard to be so quiet...')



from bonobo.util import get_name


class ApiHelper:
    # TODO __all__ kwarg only
    def __init__(self, __all__):
        self.__all__ = __all__

    def register(self, x, graph=False):
        """Register a function as being part of an API, then returns the original function."""

        if graph:
            # This function must comply to the "graph" API interface, meaning it can bahave like bonobo.run.
            from inspect import signature
            parameters = list(signature(x).parameters)
            required_parameters = {'plugins', 'services', 'strategy'}
            assert len(parameters) > 0 and parameters[
                0] == 'graph', 'First parameter of a graph api function must be "graph".'
            assert required_parameters.intersection(
                parameters
            ) == required_parameters, 'Graph api functions must define the following parameters: ' + ', '.join(
                sorted(required_parameters)
            )

        self.__all__.append(get_name(x))
        return x

    def register_graph(self, x):
        return self.register(x, graph=True)

    def register_group(self, *args, check=None):
        check = set(check) if check else None
        for attr in args:
            self.register(attr)
            if check:
                check.remove(get_name(attr))
        assert not (check and len(check))



from bonobo.constants import BEGIN, END
from bonobo.execution.strategies.base import Strategy


class NaiveStrategy(Strategy):
    # TODO: how to run plugins in "naive" mode ?

    def execute(self, graph, **kwargs):
        context = self.create_graph_execution_context(graph, **kwargs)
        context.write(BEGIN, (), END)

        # start
        context.start()

        # loop
        nodes = list(context.nodes)
        while len(nodes):
            for node in nodes:
                node.loop()
            nodes = list(node for node in nodes if node.alive)

        # stop
        context.stop()

        return context



import bonobo


def extract():
    yield 'foo'
    yield 'bar'
    yield 'baz'


def transform(x):
    return x.upper()


def load(x):
    print(x)


graph = bonobo.Graph(extract, transform, load)

graph.__doc__ = 'hello'

if __name__ == '__main__':
    bonobo.run(graph)



# -*- coding: utf-8 -*-
#
# copyright 2012-2014 romain dorgueil
#
# licensed under the apache license, version 2.0 (the "license");
# you may not use this file except in compliance with the license.
# you may obtain a copy of the license at
#
#     http://www.apache.org/licenses/license-2.0
#
# unless required by applicable law or agreed to in writing, software
# distributed under the license is distributed on an "as is" basis,
# without warranties or conditions of any kind, either express or implied.
# see the license for the specific language governing permissions and
# limitations under the license.
import time


class WithStatistics:
    def __init__(self, *names):
        self.statistics_names = names
        self.statistics = {name: 0 for name in names}

    def get_statistics(self, *args, **kwargs):
        return ((name, self.statistics[name]) for name in self.statistics_names)

    def get_statistics_as_string(self, *args, **kwargs):
        stats = tuple('{0}={1}'.format(name, cnt) for name, cnt in self.get_statistics(*args, **kwargs) if cnt > 0)
        return (kwargs.get('prefix', '') + ' '.join(stats)) if len(stats) else ''

    def increment(self, name, *, amount=1):
        self.statistics[name] += amount


class Timer:
    """
    Context manager used to time execution of stuff.
    """

    def __enter__(self):
        self.__start = time.time()

    def __exit__(self, type=None, value=None, traceback=None):
        # Error handling here
        self.__finish = time.time()

    @property
    def duration(self):
        return self.__finish - self.__start

    def __str__(self):
        return str(int(self.duration * 1000) / 1000.0) + 's'



import os

import bonobo
from bonobo.commands import BaseGraphCommand


class RunCommand(BaseGraphCommand):
    install = False
    handler = staticmethod(bonobo.run)

    def add_arguments(self, parser):
        super(RunCommand, self).add_arguments(parser)

        verbosity_group = parser.add_mutually_exclusive_group()
        verbosity_group.add_argument('--quiet', '-q', action='store_true')
        verbosity_group.add_argument('--verbose', '-v', action='store_true')

        parser.add_argument('--install', '-I', action='store_true')

    def parse_options(self, *, quiet=False, verbose=False, install=False, **options):
        from bonobo import settings
        settings.QUIET.set_if_true(quiet)
        settings.DEBUG.set_if_true(verbose)
        self.install = install
        return options

    def _run_path(self, file):
        # add install logic
        if self.install:
            if os.path.isdir(file):
                requirements = os.path.join(file, 'requirements.txt')
            else:
                requirements = os.path.join(os.path.dirname(file), 'requirements.txt')
            _install_requirements(requirements)

        return super()._run_path(file)

    def _run_module(self, mod):
        # install not implemented for a module, not sure it even make sense.
        if self.install:
            raise RuntimeError('--install behaviour when running a module is not defined.')

        return super()._run_module(mod)


def register_generic_run_arguments(parser, required=True):
    """
    Only there for backward compatibility with third party extensions.
    TODO: This should be deprecated (using the @deprecated decorator) in 0.7, and removed in 0.8 or 0.9.
    """
    dummy_command = BaseGraphCommand()
    dummy_command.required = required
    dummy_command.add_arguments(parser)
    return parser


def _install_requirements(requirements):
    """Install requirements given a path to requirements.txt file."""
    import importlib
    import pip

    pip.main(['install', '-r', requirements])
    # Some shenanigans to be sure everything is importable after this, especially .egg-link files which
    # are referenced in *.pth files and apparently loaded by site.py at some magic bootstrap moment of the
    # python interpreter.
    pip.utils.pkg_resources = importlib.reload(pip.utils.pkg_resources)
    import site
    importlib.reload(site)



import contextlib
import functools
import io
import os
import runpy
import sys
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from unittest.mock import patch

import pytest

from bonobo import open_fs, __main__, get_examples_path
from bonobo.commands import entrypoint
from bonobo.constants import Token
from bonobo.execution.contexts.graph import GraphExecutionContext
from bonobo.execution.contexts.node import NodeExecutionContext


@contextmanager
def optional_contextmanager(cm, *, ignore=False):
    if cm is None or ignore:
        yield
    else:
        with cm:
            yield


class FilesystemTester:
    def __init__(self, extension='txt', mode='w', *, input_data=''):
        self.extension = extension
        self.input_data = input_data
        self.mode = mode

    def get_services_for_reader(self, tmpdir):
        fs, filename = open_fs(tmpdir), 'input.' + self.extension
        with fs.open(filename, self.mode) as fp:
            fp.write(self.input_data)
        return fs, filename, {'fs': fs}

    def get_services_for_writer(self, tmpdir):
        fs, filename = open_fs(tmpdir), 'output.' + self.extension
        return fs, filename, {'fs': fs}


class QueueList(list):
    def append(self, item):
        if not isinstance(item, Token):
            super(QueueList, self).append(item)

    put = append


class BufferingContext:
    def __init__(self, buffer=None):
        if buffer is None:
            buffer = QueueList()
        self.buffer = buffer

    def get_buffer(self):
        return self.buffer

    def get_buffer_args_as_dicts(self):
        return [row._asdict() if hasattr(row, '_asdict') else dict(row) for row in self.buffer]


class BufferingNodeExecutionContext(BufferingContext, NodeExecutionContext):
    def __init__(self, *args, buffer=None, **kwargs):
        BufferingContext.__init__(self, buffer)
        NodeExecutionContext.__init__(self, *args, **kwargs, _outputs=[self.buffer])


class BufferingGraphExecutionContext(BufferingContext, GraphExecutionContext):
    NodeExecutionContextType = BufferingNodeExecutionContext

    def __init__(self, *args, buffer=None, **kwargs):
        BufferingContext.__init__(self, buffer)
        GraphExecutionContext.__init__(self, *args, **kwargs)

    def create_node_execution_context_for(self, node):
        return self.NodeExecutionContextType(node, parent=self, buffer=self.buffer)


def runner(f):
    @functools.wraps(f)
    def wrapped_runner(*args, catch_errors=False):
        with redirect_stdout(io.StringIO()) as stdout, redirect_stderr(io.StringIO()) as stderr:
            try:
                f(list(args))
            except BaseException as exc:
                if not catch_errors:
                    raise
                elif isinstance(catch_errors, BaseException) and not isinstance(exc, catch_errors):
                    raise
                return stdout.getvalue(), stderr.getvalue(), exc
        return stdout.getvalue(), stderr.getvalue()

    return wrapped_runner


@runner
def runner_entrypoint(args):
    """ Run bonobo using the python command entrypoint directly (bonobo.commands.entrypoint). """
    return entrypoint(args)


@runner
def runner_module(args):
    """ Run bonobo using the bonobo.__main__ file, which is equivalent as doing "python -m bonobo ..."."""
    with patch.object(sys, 'argv', ['bonobo', *args]):
        return runpy.run_path(__main__.__file__, run_name='__main__')


all_runners = pytest.mark.parametrize('runner', [runner_entrypoint, runner_module])
all_environ_targets = pytest.mark.parametrize(
    'target', [
        (get_examples_path('environ.py'), ),
        (
            '-m',
            'bonobo.examples.environ',
        ),
    ]
)


@all_runners
@all_environ_targets
class EnvironmentTestCase():
    def run_quiet(self, runner, *args):
        return runner('run', '--quiet', *args)

    def run_environ(self, runner, *args, environ=None):
        _environ = {'PATH': '/usr/bin'}
        if environ:
            _environ.update(environ)

        with patch.dict('os.environ', _environ, clear=True):
            out, err = self.run_quiet(runner, *args)
            assert 'SECRET' not in os.environ
            assert 'PASSWORD' not in os.environ
            if 'PATH' in _environ:
                assert 'PATH' in os.environ
                assert os.environ['PATH'] == _environ['PATH']

        assert err == ''
        return dict(map(lambda line: line.split(' ', 1), filter(None, out.split('\n'))))


class StaticNodeTest:
    node = None
    services = {}

    NodeExecutionContextType = BufferingNodeExecutionContext

    @contextlib.contextmanager
    def execute(self, *args, **kwargs):
        with self.NodeExecutionContextType(type(self).node, services=self.services) as context:
            yield context

    def call(self, *args, **kwargs):
        return type(self).node(*args, **kwargs)


class ConfigurableNodeTest:
    NodeType = None
    NodeExecutionContextType = BufferingNodeExecutionContext

    services = {}

    @staticmethod
    def incontext(*create_args, **create_kwargs):
        def decorator(method):
            @functools.wraps(method)
            def _incontext(self, *args, **kwargs):
                nonlocal create_args, create_kwargs
                with self.execute(*create_args, **create_kwargs) as context:
                    return method(self, context, *args, **kwargs)

            return _incontext

        return decorator

    def create(self, *args, **kwargs):
        return self.NodeType(*self.get_create_args(*args), **self.get_create_kwargs(**kwargs))

    @contextlib.contextmanager
    def execute(self, *args, **kwargs):
        with self.NodeExecutionContextType(self.create(*args, **kwargs), services=self.services) as context:
            yield context

    def get_create_args(self, *args):
        return args

    def get_create_kwargs(self, **kwargs):
        return kwargs

    def get_filesystem_tester(self):
        return FilesystemTester(self.extension, input_data=self.input_data)


class ReaderTest(ConfigurableNodeTest):
    """ Helper class to test reader transformations. """

    ReaderNodeType = None

    extension = 'txt'
    input_data = ''

    @property
    def NodeType(self):
        return self.ReaderNodeType

    @pytest.fixture(autouse=True)
    def _reader_test_fixture(self, tmpdir):
        fs_tester = self.get_filesystem_tester()
        self.fs, self.filename, self.services = fs_tester.get_services_for_reader(tmpdir)
        self.tmpdir = tmpdir

    def get_create_args(self, *args):
        return (self.filename, ) + args

    def test_customizable_output_type_transform_not_a_type(self):
        context = self.NodeExecutionContextType(
            self.create(*self.get_create_args(), output_type=str.upper, **self.get_create_kwargs()),
            services=self.services
        )
        with pytest.raises(TypeError):
            context.start()

    def test_customizable_output_type_transform_not_a_tuple(self):
        context = self.NodeExecutionContextType(
            self.create(
                *self.get_create_args(), output_type=type('UpperString', (str, ), {}), **self.get_create_kwargs()
            ),
            services=self.services
        )
        with pytest.raises(TypeError):
            context.start()


class WriterTest(ConfigurableNodeTest):
    """ Helper class to test writer transformations. """

    WriterNodeType = None

    extension = 'txt'
    input_data = ''

    @property
    def NodeType(self):
        return self.WriterNodeType

    @pytest.fixture(autouse=True)
    def _writer_test_fixture(self, tmpdir):
        fs_tester = self.get_filesystem_tester()
        self.fs, self.filename, self.services = fs_tester.get_services_for_writer(tmpdir)
        self.tmpdir = tmpdir

    def get_create_args(self, *args):
        return (self.filename, ) + args

    def readlines(self):
        with self.fs.open(self.filename) as fp:
            return tuple(map(str.strip, fp.readlines()))



import argparse
import codecs
import os
import re
import warnings
from contextlib import contextmanager

__escape_decoder = codecs.getdecoder('unicode_escape')
__posix_variable = re.compile('\$\{[^\}]*\}')


def parse_var(var):
    name, value = var.split('=', 1)

    def decode_escaped(escaped):
        return __escape_decoder(escaped)[0]

    if len(value) > 1:
        c = value[0]

        if c in ['"', "'"] and value[-1] == c:
            value = decode_escaped(value[1:-1])

    return name, value


def load_env_from_file(filename):
    """
    Read an env file into a collection of (name, value) tuples.
    """
    if not os.path.exists(filename):
        raise FileNotFoundError('Environment file {} does not exist.'.format(filename))

    with open(filename) as f:
        for lineno, line in enumerate(f):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            if '=' not in line:
                raise SyntaxError('Invalid environment file syntax in {} at line {}.'.format(filename, lineno + 1))

            name, value = parse_var(line)

            yield name, value


_parser = None


def get_argument_parser(parser=None):
    """
    Creates an argument parser with arguments to override the system environment.

    :api: bonobo.get_argument_parser

    :param _parser:
    :return:
    """
    if parser is None:
        import argparse
        parser = argparse.ArgumentParser()

    # Store globally to be able to warn the user about the fact he's probably wrong not to pass a parser to
    # parse_args(), later.
    global _parser
    _parser = parser

    _parser.add_argument('--default-env-file', '-E', action='append')
    _parser.add_argument('--default-env', action='append')
    _parser.add_argument('--env-file', action='append')
    _parser.add_argument('--env', '-e', action='append')

    return _parser


@contextmanager
def parse_args(mixed=None):
    """
    Context manager to extract and apply environment related options from the provided argparser result.

    A dictionnary with unknown options will be yielded, so the remaining options can be used by the caller.

    :api: bonobo.patch_environ

    :param mixed: ArgumentParser instance, Namespace, or dict.
    :return:
    """

    if mixed is None:
        global _parser
        if _parser is not None:
            warnings.warn(
                'You are calling bonobo.parse_args() without a parser argument, but it looks like you created a parser before. You probably want to pass your parser to this call, or if creating a new parser here is really what you want to do, please create a new one explicitely to silence this warning.'
            )
        # use the api from bonobo namespace, in case a command patched it.
        import bonobo
        mixed = bonobo.get_argument_parser()

    if isinstance(mixed, argparse.ArgumentParser):
        options = mixed.parse_args()
    else:
        options = mixed

    if not isinstance(options, dict):
        options = options.__dict__

    # make a copy so we don't polute our parent variables.
    options = dict(options)

    # storage for values before patch.
    _backup = {}

    # Priority order: --env > --env-file > system > --default-env > --default-env-file
    #
    # * The code below is reading default-env before default-env-file as if the first sets something, default-env-file
    #   won't override it.
    # * Then, env-file is read from before env, as the behaviour will be the oposite (env will override a var even if
    #   env-file sets something.)
    try:
        # Set default environment
        for name, value in map(parse_var, options.pop('default_env', []) or []):
            if not name in os.environ:
                if not name in _backup:
                    _backup[name] = os.environ.get(name, None)
                os.environ[name] = value

        # Read and set default environment from file(s)
        for filename in options.pop('default_env_file', []) or []:
            for name, value in load_env_from_file(filename):
                if not name in os.environ:
                    if not name in _backup:
                        _backup[name] = os.environ.get(name, None)
                    os.environ[name] = value

        # Read and set environment from file(s)
        for filename in options.pop('env_file', []) or []:
            for name, value in load_env_from_file(filename):
                if not name in _backup:
                    _backup[name] = os.environ.get(name, None)
                os.environ[name] = value

        # Set environment
        for name, value in map(parse_var, options.pop('env', []) or []):
            if not name in _backup:
                _backup[name] = os.environ.get(name, None)
            os.environ[name] = value

        yield options
    finally:
        for name, value in _backup.items():
            if value is None:
                del os.environ[name]
            else:
                os.environ[name] = value


@contextmanager
def change_working_directory(path):
    old_dir = os.getcwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(old_dir)



import bonobo
from bonobo.commands import BaseCommand
from bonobo.registry import READER, WRITER, default_registry
from bonobo.util.resolvers import _resolve_transformations, _resolve_options


class ConvertCommand(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('input_filename', help='Input filename.')
        parser.add_argument('output_filename', help='Output filename.')
        parser.add_argument(
            '--' + READER,
            '-r',
            help='Choose the reader factory if it cannot be detected from extension, or if detection is wrong.'
        )
        parser.add_argument(
            '--' + WRITER,
            '-w',
            help=
            'Choose the writer factory if it cannot be detected from extension, or if detection is wrong (use - for console pretty print).'
        )
        parser.add_argument(
            '--limit',
            '-l',
            type=int,
            help='Adds a Limit() after the reader instance.',
            default=None,
        )
        parser.add_argument(
            '--transformation',
            '-t',
            dest='transformation',
            action='append',
            help='Add a transformation between input and output (can be used multiple times, order is preserved).',
        )
        parser.add_argument(
            '--option',
            '-O',
            dest='option',
            action='append',
            help='Add a named option to both reader and writer factories (i.e. foo="bar").',
        )
        parser.add_argument(
            '--' + READER + '-option',
            '-' + READER[0].upper(),
            dest=READER + '_option',
            action='append',
            help='Add a named option to the reader factory.',
        )
        parser.add_argument(
            '--' + WRITER + '-option',
            '-' + WRITER[0].upper(),
            dest=WRITER + '_option',
            action='append',
            help='Add a named option to the writer factory.',
        )

    def handle(
            self,
            input_filename,
            output_filename,
            reader=None,
            reader_option=None,
            writer=None,
            writer_option=None,
            option=None,
            limit=None,
            transformation=None,
    ):
        reader_factory = default_registry.get_reader_factory_for(input_filename, format=reader)
        reader_kwargs = _resolve_options((option or []) + (reader_option or []))

        if output_filename == '-':
            writer_factory = bonobo.PrettyPrinter
            writer_args = ()
        else:
            writer_factory = default_registry.get_writer_factory_for(output_filename, format=writer)
            writer_args = (output_filename, )
        writer_kwargs = _resolve_options((option or []) + (writer_option or []))

        transformations = ()

        if limit:
            transformations += (bonobo.Limit(limit), )

        transformations += _resolve_transformations(transformation)

        graph = bonobo.Graph()
        graph.add_chain(
            reader_factory(input_filename, **reader_kwargs),
            *transformations,
            writer_factory(*writer_args, **writer_kwargs),
        )

        return bonobo.run(
            graph, services={
                'fs': bonobo.open_fs(),
            }
        )



import bonobo


def generate_data():
    yield 'foo'
    yield 'bar'
    yield 'baz'


def uppercase(x: str):
    return x.upper()


def output(x: str):
    print(x)


graph = bonobo.Graph(
    generate_data,
    uppercase,
    output,
)

if __name__ == '__main__':
    bonobo.run(graph)



import functools
import re
import sys
from keyword import iskeyword

from slugify import slugify

_class_template = '''\
from builtins import property as _property, tuple as _tuple
from operator import itemgetter as _itemgetter
from collections import OrderedDict

class {typename}(tuple):
    '{typename}({arg_list})'

    __slots__ = ()
    _attrs = {attrs!r}
    _fields = {fields!r}
    
    def __new__(_cls, {arg_list}):
        """
        Create new instance of {typename}({arg_list})
        
        """
        return _tuple.__new__(_cls, ({arg_list}))
        
    def __getnewargs__(self):
        """
        Return self as a plain tuple.
        Used by copy and pickle.
        
        """
        return tuple(self)
        
    def __repr__(self):
        """
        Return a nicely formatted representation string
        
        """
        return self.__class__.__name__ + '({repr_fmt})' % self
        
    def get(self, field, default=None):
        try:
            index = self._fields.index(field)
        except ValueError:
            return default
        return self[index]
        
    @classmethod
    def _make(cls, iterable, new=tuple.__new__, len=len):
        'Make a new {typename} object from a sequence or iterable'
        result = new(cls, iterable)
        if len(result) != {num_fields:d}:
            raise TypeError('Expected {num_fields:d} arguments, got %d' % len(result))
        return result
        
    def _replace(_self, **kwds):
        'Return a new {typename} object replacing specified fields with new values'
        result = _self._make(map(kwds.pop, {fields!r}, _self))
        if kwds:
            raise ValueError('Got unexpected field names: %r' % list(kwds))
        return result
        
    def _asdict(self):
        """
        Return a new OrderedDict which maps field names to their values.
        
        """
        return OrderedDict(zip(self._fields, self))
        
{field_defs}
'''

_field_template = '''\
    {name} = _property(_itemgetter({index:d}), doc={doc!r})
'''.strip('\n')

_reserved = frozenset(
    ['_', '_cls', '_attrs', '_fields', 'get', '_asdict', '_replace', '_make', 'self', '_self', 'tuple'] + dir(tuple)
)

_multiple_underscores_pattern = re.compile('__+')
_slugify_allowed_chars_pattern = re.compile(r'[^a-z0-9_]+', flags=re.IGNORECASE)


def _uniquify(f):
    seen = set(_reserved)

    @functools.wraps(f)
    def _uniquified(x):
        nonlocal f, seen
        x = str(x)
        v = v0 = _multiple_underscores_pattern.sub('_', f(x))
        i = 0
        # if last character is not "allowed", let's start appending indexes right from the first iteration
        if len(x) and _slugify_allowed_chars_pattern.match(x[-1]):
            v = '{}{}'.format(v0, i)
        while v in seen:
            v = '{}{}'.format(v0, i)
            i += 1
        seen.add(v)
        return v

    return _uniquified


def _make_valid_attr_name(x):
    if iskeyword(x):
        x = '_' + x
    if x.isidentifier():
        return x
    x = slugify(x, separator='_', regex_pattern=_slugify_allowed_chars_pattern)
    if x.isidentifier():
        return x
    x = '_' + x
    if x.isidentifier():
        return x
    raise ValueError(x)


def BagType(typename, fields, *, verbose=False, module=None):
    # Validate the field names.  At the user's option, either generate an error
    # message or automatically replace the field name with a valid name.

    attrs = tuple(map(_uniquify(_make_valid_attr_name), fields))
    if type(fields) is str:
        raise TypeError('BagType does not support providing fields as a string.')
    fields = list(map(str, fields))
    typename = str(typename)

    for i, name in enumerate([typename] + fields):
        if type(name) is not str:
            raise TypeError('Type names and field names must be strings, got {name!r}'.format(name=name))
        if not i:
            if not name.isidentifier():
                raise ValueError('Type names must be valid identifiers: {name!r}'.format(name=name))
            if iskeyword(name):
                raise ValueError('Type names cannot be a keyword: {name!r}'.format(name=name))

    seen = set()
    for name in fields:
        if name in seen:
            raise ValueError('Encountered duplicate field name: {name!r}'.format(name=name))
        seen.add(name)

    # Fill-in the class template
    class_definition = _class_template.format(
        typename=typename,
        fields=tuple(fields),
        attrs=attrs,
        num_fields=len(fields),
        arg_list=repr(attrs).replace("'", "")[1:-1],
        repr_fmt=', '.join(('%r' if isinstance(fields[index], int) else '{name}=%r').format(name=name)
                           for index, name in enumerate(attrs)),
        field_defs='\n'.join(
            _field_template.format(
                index=index,
                name=name,
                doc='Alias for ' +
                ('field #{}'.format(index) if isinstance(fields[index], int) else repr(fields[index]))
            ) for index, name in enumerate(attrs)
        )
    )

    # Execute the template string in a temporary namespace and support
    # tracing utilities by setting a value for frame.f_globals['__name__']
    namespace = dict(__name__='namedtuple_%s' % typename)
    exec(class_definition, namespace)
    result = namespace[typename]
    result._source = class_definition
    if verbose:
        print(result._source)

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in environments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython), or where the user has
    # specified a particular module.
    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass
    if module is not None:
        result.__module__ = module

    return result



from bonobo.plugins.jupyter import JupyterOutputPlugin


def _jupyter_nbextension_paths():
    return [{'section': 'notebook', 'src': 'static', 'dest': 'bonobo-jupyter', 'require': 'bonobo-jupyter/extension'}]


__all__ = [
    'JupyterOutputPlugin',
]



import mimetypes

import os

from bonobo import JsonReader, CsvReader, PickleReader, FileReader, FileWriter, PickleWriter, CsvWriter, JsonWriter

FILETYPE_CSV = 'text/csv'
FILETYPE_JSON = 'application/json'
FILETYPE_PICKLE = 'pickle'
FILETYPE_PLAIN = 'text/plain'

READER = 'reader'
WRITER = 'writer'


class Registry:
    ALIASES = {
        'csv': FILETYPE_CSV,
        'json': FILETYPE_JSON,
        'pickle': FILETYPE_PICKLE,
        'plain': FILETYPE_PLAIN,
        'text': FILETYPE_PLAIN,
        'txt': FILETYPE_PLAIN,
    }

    FACTORIES = {
        READER: {
            FILETYPE_JSON: JsonReader,
            FILETYPE_CSV: CsvReader,
            FILETYPE_PICKLE: PickleReader,
            FILETYPE_PLAIN: FileReader,
        },
        WRITER: {
            FILETYPE_JSON: JsonWriter,
            FILETYPE_CSV: CsvWriter,
            FILETYPE_PICKLE: PickleWriter,
            FILETYPE_PLAIN: FileWriter,
        },
    }

    def get_factory_for(self, kind, name, *, format=None):
        if not kind in self.FACTORIES:
            raise KeyError('Unknown factory kind {!r}.'.format(kind))

        if format is None and name is None:
            raise RuntimeError('Cannot guess factory without at least a filename or a format.')

        # Guess mimetype if possible
        if format is None:
            format = mimetypes.guess_type(name)[0]

        # Guess from extension if possible
        if format is None:
            _, _ext = os.path.splitext(name)
            if _ext:
                format = _ext[1:]

        # Apply aliases
        if format in self.ALIASES:
            format = self.ALIASES[format]

        if format is None or not format in self.FACTORIES[kind]:
            raise RuntimeError(
                'Could not resolve {kind} factory for {name} ({format}).'.format(kind=kind, name=name, format=format)
            )

        return self.FACTORIES[kind][format]

    def get_reader_factory_for(self, name, *, format=None):
        """
        Returns a callable to build a reader for the provided filename, eventually forcing a format.

        :param name: filename
        :param format: format
        :return: type
        """
        return self.get_factory_for(READER, name, format=format)

    def get_writer_factory_for(self, name, *, format=None):
        """
        Returns a callable to build a writer for the provided filename, eventually forcing a format.

        :param name: filename
        :param format: format
        :return: type
        """
        return self.get_factory_for(WRITER, name, format=format)


default_registry = Registry()



from logging import getLogger
from types import GeneratorType

import bonobo
from bonobo.plugins.console import ConsoleOutputPlugin
from bonobo.util.term import CLEAR_EOL
from colorama import Fore, Back, Style
from django.core.management import BaseCommand
from django.core.management.base import OutputWrapper
from mondrian import term

from .utils import create_or_update


class ETLCommand(BaseCommand):
    @property
    def logger(self):
        try:
            return self._logger
        except AttributeError:
            self._logger = getLogger(type(self).__module__)
            return self._logger

    create_or_update = staticmethod(create_or_update)

    def create_parser(self, prog_name, subcommand):
        return bonobo.get_argument_parser(super().create_parser(prog_name, subcommand))

    def get_graph(self, *args, **options):
        def not_implemented():
            raise NotImplementedError('You must implement {}.get_graph() method.'.format(self))

        return bonobo.Graph(not_implemented)

    def get_services(self):
        return {}

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def handle(self, *args, **options):
        _stdout_backup, _stderr_backup = self.stdout, self.stderr

        self.stdout = OutputWrapper(ConsoleOutputPlugin._stdout, ending=CLEAR_EOL + '\n')
        self.stderr = OutputWrapper(ConsoleOutputPlugin._stderr, ending=CLEAR_EOL + '\n')
        self.stderr.style_func = lambda x: Fore.LIGHTRED_EX + Back.RED + '!' + Style.RESET_ALL + ' ' + x

        with bonobo.parse_args(options) as options:
            services = self.get_services()
            graph_coll = self.get_graph(*args, **options)

            if not isinstance(graph_coll, GeneratorType):
                graph_coll = (graph_coll,)

            for i, graph in enumerate(graph_coll):
                assert isinstance(graph, bonobo.Graph), 'Invalid graph provided.'
                print(term.lightwhite('{}. {}'.format(i + 1, graph.name)))
                result = bonobo.run(graph, services=services)
                print(term.lightblack(' ... return value: ' + str(result)))
                print()

        self.stdout, self.stderr = _stdout_backup, _stderr_backup



import bonobo
import datetime
import time


def extract():
    """Placeholder, change, rename, remove... """
    for x in range(60):
        if x:
            time.sleep(1)
        yield datetime.datetime.now()


def get_graph():
    graph = bonobo.Graph()
    graph.add_chain(
        extract,
        print,
    )

    return graph


if __name__ == '__main__':
    parser = bonobo.get_argument_parser()
    with bonobo.parse_args(parser):
        bonobo.run(get_graph())



import threading
import time

from bonobo.config import Configurable, ContextProcessor, Method, Option


class RateLimitBucket(threading.Thread):
    daemon = True

    @property
    def stopped(self):
        return self._stop_event.is_set()

    def __init__(self, initial=1, period=1, amount=1):
        super(RateLimitBucket, self).__init__()
        self.semaphore = threading.BoundedSemaphore(initial)
        self.amount = amount
        self.period = period

        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def run(self):
        while not self.stopped:
            time.sleep(self.period)
            for _ in range(self.amount):
                self.semaphore.release()

    def wait(self):
        return self.semaphore.acquire()


class RateLimited(Configurable):
    handler = Method()

    initial = Option(int, positional=True, default=1)
    period = Option(int, positional=True, default=1)
    amount = Option(int, positional=True, default=1)

    @ContextProcessor
    def bucket(self, context):
        bucket = RateLimitBucket(self.initial, self.amount, self.period)
        bucket.start()
        yield bucket
        bucket.stop()
        bucket.join()

    def __call__(self, bucket, *args, **kwargs):
        bucket.wait()
        return self.handler(*args, **kwargs)



""" Readers and writers for common file formats. """

from .csv import CsvReader, CsvWriter
from .file import FileReader, FileWriter
from .json import JsonReader, JsonWriter, LdjsonReader, LdjsonWriter
from .pickle import PickleReader, PickleWriter

__all__ = [
    'CsvReader',
    'CsvWriter',
    'FileReader',
    'FileWriter',
    'JsonReader',
    'JsonWriter',
    'LdjsonReader',
    'LdjsonWriter',
    'PickleReader',
    'PickleWriter',
]



from whistle import Event

START = 'execution.start'
STARTED = 'execution.started'
TICK = 'execution.tick'
STOP = 'execution.stop'
STOPPED = 'execution.stopped'
KILL = 'execution.kill'


class ExecutionEvent(Event):
    def __init__(self, context):
        self.context = context



import os

from jinja2 import Environment, FileSystemLoader

from bonobo.commands import BaseCommand


class InitCommand(BaseCommand):
    TEMPLATES = {'bare', 'default'}
    TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), 'templates')

    def add_arguments(self, parser):
        parser.add_argument('filename')
        parser.add_argument('--force', '-f', default=False, action='store_true')

        target_group = parser.add_mutually_exclusive_group(required=False)
        target_group.add_argument('--template', '-t', choices=self.TEMPLATES, default='default')
        target_group.add_argument('--package', '-p', action='store_true', default=False)

    def create_file_from_template(self, *, template, filename):
        template_name = template
        name, ext = os.path.splitext(filename)
        if ext != '.py':
            raise ValueError('Filenames should end with ".py".')

        loader = FileSystemLoader(self.TEMPLATES_PATH)
        env = Environment(loader=loader)
        template = env.get_template(template_name + '.py-tpl')

        with open(filename, 'w+') as f:
            f.write(template.render(name=name))

        self.logger.info('Generated {} using template {!r}.'.format(filename, template_name))

    def create_package(self, *, filename):
        name, ext = os.path.splitext(filename)
        if ext != '':
            raise ValueError('Package names should not have an extension.')

        try:
            import medikit.commands
        except ImportError as exc:
            raise ImportError(
                'To initialize a package, you need to install medikit (pip install --upgrade medikit).'
            ) from exc

        package_name = os.path.basename(filename)
        medikit.commands.handle_init(
            os.path.join(os.getcwd(), filename, 'Projectfile'), name=package_name, requirements=['bonobo']
        )

        self.logger.info('Generated "{}" package with medikit.'.format(package_name))
        self.create_file_from_template(template='default', filename=os.path.join(filename, package_name, '__main__.py'))

        print('Your "{}" package has been created.'.format(package_name))
        print()
        print('Install it...')
        print()
        print('    pip install --editable {}'.format(filename))
        print()
        print('Then maybe run the example...')
        print()
        print('    python -m {}'.format(package_name))
        print()
        print('Enjoy!')

    def handle(self, *, template, filename, package=False, force=False):
        if os.path.exists(filename) and not force:
            raise FileExistsError('Target filename already exists, use --force to override.')

        if package:
            self.create_package(filename=filename)
        else:
            self.create_file_from_template(template=template, filename=filename)



import bonobo
from bonobo.examples.types.strings import get_graph

if __name__ == '__main__':
    parser = bonobo.get_argument_parser()
    with bonobo.parse_args(parser):
        bonobo.run(get_graph())



import logging
import sys
from collections import namedtuple
from queue import Empty
from time import sleep
from types import GeneratorType

from bonobo.config import create_container
from bonobo.config.processors import ContextCurrifier
from bonobo.constants import NOT_MODIFIED, BEGIN, END, TICK_PERIOD, Token, Flag, INHERIT
from bonobo.errors import InactiveReadableError, UnrecoverableError, UnrecoverableTypeError
from bonobo.execution.contexts.base import BaseContext
from bonobo.structs.inputs import Input
from bonobo.util import get_name, isconfigurabletype, ensure_tuple
from bonobo.util.bags import BagType
from bonobo.util.statistics import WithStatistics

logger = logging.getLogger(__name__)

UnboundArguments = namedtuple('UnboundArguments', ['args', 'kwargs'])


class NodeExecutionContext(BaseContext, WithStatistics):
    def __init__(self, wrapped, *, parent=None, services=None, _input=None, _outputs=None):
        """
        Node execution context has the responsibility fo storing the state of a transformation during its execution.

        :param wrapped: wrapped transformation
        :param parent: parent context, most probably a graph context
        :param services: dict-like collection of services
        :param _input: input queue (optional)
        :param _outputs: output queues (optional)
        """
        BaseContext.__init__(self, wrapped, parent=parent)
        WithStatistics.__init__(self, 'in', 'out', 'err', 'warn')

        # Services: how we'll access external dependencies
        if services:
            if self.parent:
                raise RuntimeError(
                    'Having services defined both in GraphExecutionContext and child NodeExecutionContext is not supported, for now.'
                )
            self.services = create_container(services)
        else:
            self.services = None

        # Input / Output: how the wrapped node will communicate
        self.input = _input or Input()
        self.outputs = _outputs or []

        # Types
        self._input_type, self._input_length = None, None
        self._output_type = None

        # Stack: context decorators for the execution
        self._stack = None

    def __str__(self):
        return self.__name__ + self.get_statistics_as_string(prefix=' ')

    def __repr__(self):
        name, type_name = get_name(self), get_name(type(self))
        return '<{}({}{}){}>'.format(type_name, self.status, name, self.get_statistics_as_string(prefix=' '))

    def start(self):
        """
        Starts this context, a.k.a the phase where you setup everything which will be necessary during the whole
        lifetime of a transformation.

        The "ContextCurrifier" is in charge of setting up a decorating stack, that includes both services and context
        processors, and will call the actual node callable with additional parameters.

        """
        super().start()

        try:
            initial = self._get_initial_context()
            self._stack = ContextCurrifier(self.wrapped, *initial.args, **initial.kwargs)
            if isconfigurabletype(self.wrapped):
                # Not normal to have a partially configured object here, so let's warn the user instead of having get into
                # the hard trouble of understanding that by himself.
                raise TypeError(
                    'Configurables should be instanciated before execution starts.\nGot {!r}.\n'.format(self.wrapped)
                )
            self._stack.setup(self)
        except Exception:
            # Set the logging level to the lowest possible, to avoid double log.
            self.fatal(sys.exc_info(), level=0)

            # We raise again, so the error is not ignored out of execution loops.
            raise

    def loop(self):
        """
        The actual infinite loop for this transformation.

        """
        logger.debug('Node loop starts for {!r}.'.format(self))

        while self.should_loop:
            try:
                self.step()
            except InactiveReadableError:
                break
            except Empty:
                sleep(TICK_PERIOD)  # XXX: How do we determine this constant?
                continue
            except (
                    NotImplementedError,
                    UnrecoverableError,
            ):
                self.fatal(sys.exc_info())  # exit loop
            except Exception:  # pylint: disable=broad-except
                self.error(sys.exc_info())  # does not exit loop
            except BaseException:
                self.fatal(sys.exc_info())  # exit loop

        logger.debug('Node loop ends for {!r}.'.format(self))

    def step(self):
        """
        A single step in the loop.

        Basically gets an input bag, send it to the node, interpret the results.

        """

        # Pull and check data
        input_bag = self._get()

        # Sent through the stack
        results = self._stack(input_bag)

        # self._exec_time += timer.duration
        # Put data onto output channels

        if isinstance(results, GeneratorType):
            while True:
                try:
                    # if kill flag was step, stop iterating.
                    if self._killed:
                        break
                    result = next(results)
                except StopIteration:
                    # That's not an error, we're just done.
                    break
                else:
                    # Push data (in case of an iterator)
                    self._send(self._cast(input_bag, result))
        elif results:
            # Push data (returned value)
            self._send(self._cast(input_bag, results))
        else:
            # case with no result, an execution went through anyway, use for stats.
            # self._exec_count += 1
            pass

    def stop(self):
        """
        Cleanup the context, after the loop ended.

        """
        if self._stack:
            try:
                self._stack.teardown()
            except:
                self.fatal(sys.exc_info())

        super().stop()

    def send(self, *_output, _input=None):
        return self._send(self._cast(_input, _output))

    ### Input type and fields
    @property
    def input_type(self):
        return self._input_type

    def set_input_type(self, input_type):
        if self._input_type is not None:
            raise RuntimeError('Cannot override input type, already have %r.', self._input_type)

        if type(input_type) is not type:
            raise UnrecoverableTypeError('Input types must be regular python types.')

        if not issubclass(input_type, tuple):
            raise UnrecoverableTypeError('Input types must be subclasses of tuple (and act as tuples).')

        self._input_type = input_type

    def get_input_fields(self):
        return self._input_type._fields if self._input_type and hasattr(self._input_type, '_fields') else None

    def set_input_fields(self, fields, typename='Bag'):
        self.set_input_type(BagType(typename, fields))

    ### Output type and fields
    @property
    def output_type(self):
        return self._output_type

    def set_output_type(self, output_type):
        if self._output_type is not None:
            raise RuntimeError('Cannot override output type, already have %r.', self._output_type)

        if type(output_type) is not type:
            raise UnrecoverableTypeError('Output types must be regular python types.')

        if not issubclass(output_type, tuple):
            raise UnrecoverableTypeError('Output types must be subclasses of tuple (and act as tuples).')

        self._output_type = output_type

    def get_output_fields(self):
        return self._output_type._fields if self._output_type and hasattr(self._output_type, '_fields') else None

    def set_output_fields(self, fields, typename='Bag'):
        self.set_output_type(BagType(typename, fields))

    ### Attributes
    def setdefault(self, attr, value):
        try:
            getattr(self, attr)
        except AttributeError:
            setattr(self, attr, value)

    def write(self, *messages):
        """
        Push a message list to this context's input queue.

        :param mixed value: message
        """
        for message in messages:
            if isinstance(message, Token):
                self.input.put(message)
            elif self._input_type:
                self.input.put(ensure_tuple(message, cls=self._input_type))
            else:
                self.input.put(ensure_tuple(message))

    def write_sync(self, *messages):
        self.write(BEGIN, *messages, END)
        for _ in messages:
            self.step()

    def error(self, exc_info, *, level=logging.ERROR):
        self.increment('err')
        super().error(exc_info, level=level)

    def fatal(self, exc_info, *, level=logging.CRITICAL):
        self.increment('err')
        super().fatal(exc_info, level=level)
        self.input.shutdown()

    def get_service(self, name):
        if self.parent:
            return self.parent.services.get(name)
        return self.services.get(name)

    def _get(self):
        """
        Read from the input queue.

        If Queue raises (like Timeout or Empty), stat won't be changed.

        """
        input_bag = self.input.get()

        # Store or check input type
        if self._input_type is None:
            self._input_type = type(input_bag)
        elif type(input_bag) is not self._input_type:
            raise UnrecoverableTypeError(
                'Input type changed between calls to {!r}.\nGot {!r} which is not of type {!r}.'.format(
                    self.wrapped, input_bag, self._input_type
                )
            )

        # Store or check input length, which is a soft fallback in case we're just using tuples
        if self._input_length is None:
            self._input_length = len(input_bag)
        elif len(input_bag) != self._input_length:
            raise UnrecoverableTypeError(
                'Input length changed between calls to {!r}.\nExpected {} but got {}: {!r}.'.format(
                    self.wrapped, self._input_length, len(input_bag), input_bag
                )
            )

        self.increment('in')  # XXX should that go before type check ?

        return input_bag

    def _cast(self, _input, _output):
        """
        Transforms a pair of input/output into the real slim output.

        :param _input: Bag
        :param _output: mixed
        :return: Bag
        """

        tokens, _output = split_token(_output)

        if NOT_MODIFIED in tokens:
            return ensure_tuple(_input, cls=(self.output_type or tuple))

        if INHERIT in tokens:
            if self._output_type is None:
                self._output_type = concat_types(self._input_type, self._input_length, self._output_type, len(_output))
            _output = _input + ensure_tuple(_output)

        return ensure_tuple(_output, cls=(self._output_type or tuple))

    def _send(self, value, _control=False):
        """
        Sends a message to all of this context's outputs.

        :param mixed value: message
        :param _control: if true, won't count in statistics.
        """

        if not _control:
            self.increment('out')

        for output in self.outputs:
            output.put(value)

    def _get_initial_context(self):
        if self.parent:
            return UnboundArguments((), self.parent.services.kwargs_for(self.wrapped))
        if self.services:
            return UnboundArguments((), self.services.kwargs_for(self.wrapped))
        return UnboundArguments((), {})


def isflag(param):
    return isinstance(param, Flag)


def split_token(output):
    """
    Split an output into token tuple, real output tuple.

    :param output:
    :return: tuple, tuple
    """

    output = ensure_tuple(output)

    flags, i, len_output, data_allowed = set(), 0, len(output), True
    while i < len_output and isflag(output[i]):
        if output[i].must_be_first and i:
            raise ValueError('{} flag must be first.'.format(output[i]))
        if i and output[i - 1].must_be_last:
            raise ValueError('{} flag must be last.'.format(output[i - 1]))
        if output[i] in flags:
            raise ValueError('Duplicate flag {}.'.format(output[i]))
        flags.add(output[i])
        data_allowed &= output[i].allows_data
        i += 1

    output = output[i:]
    if not data_allowed and len(output):
        raise ValueError('Output data provided after a flag that does not allow data.')
    return flags, output


def concat_types(t1, l1, t2, l2):
    t1, t2 = t1 or tuple, t2 or tuple

    if t1 == t2 == tuple:
        return tuple

    f1 = t1._fields if hasattr(t1, '_fields') else tuple(range(l1))
    f2 = t2._fields if hasattr(t2, '_fields') else tuple(range(l2))

    return BagType('Inherited', f1 + f2)



import bonobo

# Represent our data processor as a simple directed graph of callables.
graph = bonobo.Graph(
    ['foo', 'bar', 'baz'],
    str.upper,
    print,
)

if __name__ == '__main__':
    bonobo.run(graph)



class Plugin:
    """
    A plugin is an extension to the core behavior of bonobo. If you're writing transformations, you should not need
    to use this interface.
    
    For examples, you can read bonobo.plugins.console.ConsoleOutputPlugin, or bonobo.plugins.jupyter.JupyterOutputPlugin
    that respectively permits an interactive output on an ANSI console and a rich output in a jupyter notebook. Note
    that you most probably won't instanciate them by yourself at runtime, as it's the default behaviour of bonobo to use
    them if your in a compatible context (aka an interactive terminal for the console plugin, or a jupyter notebook for
    the notebook plugin.)
    
    Warning: THE PLUGIN API IS PRE-ALPHA AND WILL EVOLVE BEFORE 1.0, DO NOT RELY ON IT BEING STABLE!

    """

    def register(self, dispatcher):
        """
        :param dispatcher: whistle.EventDispatcher
        """
        pass

    def unregister(self, dispatcher):
        """
        :param dispatcher: whistle.EventDispatcher
        """
        pass



import bonobo
from bonobo import examples
from bonobo.contrib.opendatasoft import OpenDataSoftAPI as ODSReader
from bonobo.examples.datasets.services import get_services


def get_graph(graph=None, *, _limit=(), _print=()):
    graph = graph or bonobo.Graph()

    producer = graph.add_chain(
        ODSReader(
            dataset='liste-des-cafes-a-un-euro',
            netloc='opendata.paris.fr'
        ),
        *_limit,
        bonobo.UnpackItems(0),
        bonobo.Rename(
            name='nom_du_cafe',
            address='adresse',
            zipcode='arrondissement'
        ),
        bonobo.Format(city='Paris', country='France'),
        bonobo.OrderFields(
            [
                'name', 'address', 'zipcode', 'city', 'country',
                'geometry', 'geoloc'
            ]
        ),
        *_print,
    )

    # Comma separated values.
    graph.add_chain(
        bonobo.CsvWriter(
            'coffeeshops.csv',
            fields=['name', 'address', 'zipcode', 'city'],
            delimiter=','
        ),
        _input=producer.output,
    )

    # Standard JSON
    graph.add_chain(
        bonobo.JsonWriter(path='coffeeshops.json'),
        _input=producer.output,
    )

    # Line-delimited JSON
    graph.add_chain(
        bonobo.LdjsonWriter(path='coffeeshops.ldjson'),
        _input=producer.output,
    )

    return graph


if __name__ == '__main__':
    parser = examples.get_argument_parser()

    with bonobo.parse_args(parser) as options:
        bonobo.run(
            get_graph(**examples.get_graph_options(options)),
            services=get_services()
        )



from bonobo.config import Option, ContextProcessor, use_context
from bonobo.constants import NOT_MODIFIED
from bonobo.errors import UnrecoverableError
from bonobo.nodes.io.base import FileHandler, Reader, Writer
from bonobo.util import ensure_tuple


class FileReader(Reader, FileHandler):
    """Component factory for file-like readers.

    On its own, it can be used to read a file and yield one row per line, trimming the "eol" character at the end if
    present. Extending it is usually the right way to create more specific file readers (like json, csv, etc.)
    """

    mode = Option(str, default='r', __doc__='''
        What mode to use for open() call.
    ''')  # type: str

    output_fields = Option(
        ensure_tuple,
        required=False,
        __doc__='''
        Specify the field names of output lines.
        Mutually exclusive with "output_type".
    '''
    )
    output_type = Option(
        required=False,
        __doc__='''
        Specify the type of output lines.
        Mutually exclusive with "output_fields".
    '''
    )

    @ContextProcessor
    def output(self, context, *args, **kwargs):
        """
        Allow all readers to use eventually use output_fields XOR output_type options.

        """

        output_fields = self.output_fields
        output_type = self.output_type

        if output_fields and output_type:
            raise UnrecoverableError('Cannot specify both output_fields and output_type option.')

        if self.output_type:
            context.set_output_type(self.output_type)

        if self.output_fields:
            context.set_output_fields(self.output_fields)

        yield

    def read(self, file, *, fs):
        """
        Write a row on the next line of given file.
        Prefix is used for newlines.
        """
        for line in file:
            yield line.rstrip(self.eol)

    __call__ = read


@use_context
class FileWriter(Writer, FileHandler):
    """Component factory for file or file-like writers.

    On its own, it can be used to write in a file one line per row that comes into this component. Extending it is
    usually the right way to create more specific file writers (like json, csv, etc.)
    """

    mode = Option(str, default='w+', __doc__='''
        What mode to use for open() call.
    ''')  # type: str

    def write(self, file, context, line, *, fs):
        """
        Write a row on the next line of opened file in context.
        """
        context.setdefault('lineno', 0)
        self._write_line(file, (self.eol if context.lineno else '') + line)
        context.lineno += 1
        return NOT_MODIFIED

    def _write_line(self, file, line):
        return file.write(line)

    __call__ = write



import pickle

from bonobo.config import Option, use_context
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.base import FileHandler
from bonobo.nodes.io.file import FileReader, FileWriter


class PickleHandler(FileHandler):
    """

    .. attribute:: item_names

        The names of the items in the pickle, if it is not defined in the first item of the pickle.

    """

    fields = Option(tuple, required=False)


@use_context
class PickleReader(FileReader, PickleHandler):
    """
    Reads a Python pickle object and yields the items in dicts.
    """

    mode = Option(str, default='rb')

    def read(self, file, context, *, fs):
        data = pickle.load(file)

        # if the data is not iterable, then wrap the object in a list so it may be iterated
        if isinstance(data, dict):
            is_dict = True
            iterator = iter(data.items())
        else:
            is_dict = False
            try:
                iterator = iter(data)
            except TypeError:
                iterator = iter([data])

        if not context.output_type:
            context.set_output_fields(self.fields or next(iterator))
        fields = context.get_output_fields()
        fields_length = len(fields)

        for row in iterator:
            if len(row) != fields_length:
                raise ValueError('Received an object with {} items, expected {}.'.format(len(row), fields_length))

            yield tuple(row.values() if is_dict else row)

    __call__ = read


@use_context
class PickleWriter(FileWriter, PickleHandler):
    mode = Option(str, default='wb')

    def write(self, file, context, item, *, fs):
        """
        Write a pickled item to the opened file.
        """
        context.setdefault('lineno', 0)
        file.write(pickle.dumps(item))
        context.lineno += 1
        return NOT_MODIFIED

    __call__ = write



import bonobo
from bonobo import examples
from bonobo.examples.files._services import get_services


def skip_comments(line):
    line = line.strip()
    if not line.startswith('#'):
        yield line


def get_graph(*, _limit=(), _print=()):
    return bonobo.Graph(
        bonobo.FileReader('datasets/passwd.txt'),
        skip_comments,
        *_limit,
        lambda s: s.split(':')[0],
        *_print,
        bonobo.FileWriter('usernames.txt', fs='fs.output'),
    )


if __name__ == '__main__':
    parser = examples.get_argument_parser()
    with bonobo.parse_args(parser) as options:
        bonobo.run(
            get_graph(**examples.get_graph_options(options)),
            services=get_services()
        )



from bonobo.execution.contexts.base import BaseContext


class PluginExecutionContext(BaseContext):
    @property
    def dispatcher(self):
        return self.parent.dispatcher

    def register(self):
        return self.wrapped.register(self.dispatcher)

    def unregister(self):
        return self.wrapped.unregister(self.dispatcher)



import bonobo
from bonobo.examples.files._services import get_services


def get_graph(*, _limit=None, _print=False):
    return bonobo.Graph(
        bonobo.CsvReader('datasets/coffeeshops.txt'),
        *((bonobo.Limit(_limit), ) if _limit else ()),
        *((bonobo.PrettyPrinter(), ) if _print else ()),
        bonobo.CsvWriter('coffeeshops.csv', fs='fs.output')
    )


if __name__ == '__main__':
    parser = bonobo.get_argument_parser()

    parser.add_argument(
        '--limit',
        '-l',
        type=int,
        default=None,
        help='If set, limits the number of processed lines.'
    )
    parser.add_argument(
        '--print',
        '-p',
        action='store_true',
        default=False,
        help='If set, pretty prints before writing to output file.'
    )

    with bonobo.parse_args(parser) as options:
        bonobo.run(
            get_graph(_limit=options['limit'], _print=options['print']),
            services=get_services()
        )



import io
import sys
from contextlib import redirect_stdout, redirect_stderr

from colorama import Style, Fore, init as initialize_colorama_output_wrappers

from bonobo import settings
from bonobo.execution import events
from bonobo.plugins import Plugin
from bonobo.util.term import CLEAR_EOL, MOVE_CURSOR_UP

initialize_colorama_output_wrappers(wrap=True)


class ConsoleOutputPlugin(Plugin):
    """
    Outputs status information to the connected stdout. Can be a TTY, with or without support for colors/cursor
    movements, or a non tty (pipe, file, ...). The features are adapted to terminal capabilities.

    On Windows, we'll play a bit differently because we don't know how to manipulate cursor position. We'll only
    display stats at the very end, and there won't be this "buffering" logic we need to display both stats and stdout.

    .. attribute:: prefix

        String prefix of output lines.

    """

    # Standard outputs descriptors backup here, also used to override if needed.
    _stdout = sys.stdout
    _stderr = sys.stderr

    # When the plugin is instanciated, we'll set the real value of this.
    isatty = False

    # Whether we're on windows, or a real operating system.
    iswindows = (sys.platform == 'win32')

    def __init__(self):
        self.isatty = self._stdout.isatty()

    def register(self, dispatcher):
        dispatcher.add_listener(events.START, self.setup)
        dispatcher.add_listener(events.TICK, self.tick)
        dispatcher.add_listener(events.STOPPED, self.teardown)

    def unregister(self, dispatcher):
        dispatcher.remove_listener(events.STOPPED, self.teardown)
        dispatcher.remove_listener(events.TICK, self.tick)
        dispatcher.remove_listener(events.START, self.setup)

    def setup(self, event):
        # TODO this wont work if one instance is registered with more than one context.
        # Two options:
        # - move state to context
        # - forbid registering more than once
        self.prefix = ''
        self.counter = 0
        self._append_cache = ''

        self.stdout = IOBuffer()
        self.redirect_stdout = redirect_stdout(self._stdout if self.iswindows else self.stdout)
        self.redirect_stdout.__enter__()

        self.stderr = IOBuffer()
        self.redirect_stderr = redirect_stderr(self._stderr if self.iswindows else self.stderr)
        self.redirect_stderr.__enter__()

    def tick(self, event):
        if self.isatty and not self.iswindows:
            self._write(event.context, rewind=True)
        else:
            pass  # not a tty, or windows, so we'll ignore stats output

    def teardown(self, event):
        self._write(event.context, rewind=False)
        self.redirect_stderr.__exit__(None, None, None)
        self.redirect_stdout.__exit__(None, None, None)

    def write(self, context, prefix='', rewind=True, append=None):
        t_cnt = len(context)

        if not self.iswindows:
            for line in self.stdout.switch().split('\n')[:-1]:
                print(line + CLEAR_EOL, file=self._stdout)
            for line in self.stderr.switch().split('\n')[:-1]:
                print(line + CLEAR_EOL, file=self._stderr)

        alive_color = Style.BRIGHT
        dead_color = Style.BRIGHT + Fore.BLACK

        for i in context.graph.topologically_sorted_indexes:
            node = context[i]
            name_suffix = '({})'.format(i) if settings.DEBUG.get() else ''

            liveliness_color = alive_color if node.alive else dead_color
            liveliness_prefix = ' {}{}{} '.format(liveliness_color, node.status, Style.RESET_ALL)
            _line = ''.join((
                liveliness_prefix,
                node.name,
                name_suffix,
                ' ',
                node.get_statistics_as_string(),
                ' ',
                node.get_flags_as_string(),
                Style.RESET_ALL,
                ' ',
            ))
            print(prefix + _line + CLEAR_EOL, file=self._stderr)

        if append:
            # todo handle multiline
            print(
                ''.join((
                    ' `-> ', ' '.join('{}{}{}: {}'.format(Style.BRIGHT, k, Style.RESET_ALL, v) for k, v in append),
                    CLEAR_EOL
                )),
                file=self._stderr
            )
            t_cnt += 1

        if rewind:
            print(CLEAR_EOL, file=self._stderr)
            print(MOVE_CURSOR_UP(t_cnt + 2), file=self._stderr)

    def _write(self, context, rewind):
        if settings.PROFILE.get():
            if self.counter % 10 and self._append_cache:
                append = self._append_cache
            else:
                self._append_cache = append = (('Memory', '{0:.2f} Mb'.format(memory_usage())),
                                               # ('Total time', '{0} s'.format(execution_time(harness))),
                                               )
        else:
            append = ()
        self.write(context, prefix=self.prefix, append=append, rewind=rewind)
        self.counter += 1


class IOBuffer():
    """
    The role of IOBuffer is to overcome the problem of multiple threads wanting to write to stdout at the same time. It
    works a bit like a videogame: there are two buffers, one that is used to write, and one which is used to read from.
    On each cycle, we swap the buffers, and the console plugin handle output of the one which is not anymore "active".

    """

    def __init__(self):
        self.current = io.StringIO()
        self.write = self.current.write

    def switch(self):
        previous = self.current
        self.current = io.StringIO()
        self.write = self.current.write
        try:
            return previous.getvalue()
        finally:
            previous.close()

    def flush(self):
        self.current.flush()


def memory_usage():
    import os, psutil
    process = psutil.Process(os.getpid())
    return process.memory_info()[0] / float(2**20)



from bonobo.util import get_name


class InactiveIOError(IOError):
    pass


class InactiveReadableError(InactiveIOError):
    pass


class InactiveWritableError(InactiveIOError):
    pass


class ValidationError(RuntimeError):
    def __init__(self, inst, message):
        super(ValidationError, self).__init__(
            'Validation error in {class_name}: {message}'.format(
                class_name=type(inst).__name__,
                message=message,
            )
        )


class ProhibitedOperationError(RuntimeError):
    pass


class ConfigurationError(Exception):
    pass


class UnrecoverableError(Exception):
    """Flag for errors that must interrupt the workflow, either because they will happen for sure on each node run, or
    because you know that your transformation has no point continuing runnning after a bad event."""


class AbstractError(UnrecoverableError, NotImplementedError):
    """Abstract error is a convenient error to declare a method as "being left as an exercise for the reader"."""

    def __init__(self, method):
        super().__init__(
            'Call to abstract method {class_name}.{method_name}(...): missing implementation.'.format(
                class_name=get_name(method.__self__),
                method_name=get_name(method),
            )
        )


class UnrecoverableTypeError(UnrecoverableError, TypeError):
    pass


class UnrecoverableValueError(UnrecoverableError, ValueError):
    pass


class UnrecoverableRuntimeError(UnrecoverableError, RuntimeError):
    pass


class UnrecoverableNotImplementedError(UnrecoverableError, NotImplementedError):
    pass


class MissingServiceImplementationError(UnrecoverableError, KeyError):
    pass



import os

import bonobo
from bonobo import examples
from bonobo.examples.datasets.coffeeshops import get_graph as get_coffeeshops_graph
from bonobo.examples.datasets.fablabs import get_graph as get_fablabs_graph
from bonobo.examples.datasets.services import get_services, get_datasets_dir, get_minor_version

graph_factories = {
    'coffeeshops': get_coffeeshops_graph,
    'fablabs': get_fablabs_graph,
}

if __name__ == '__main__':
    parser = examples.get_argument_parser()
    parser.add_argument(
        '--target', '-t', choices=graph_factories.keys(), nargs='+'
    )
    parser.add_argument('--sync', action='store_true', default=False)

    with bonobo.parse_args(parser) as options:
        graph_options = examples.get_graph_options(options)
        graph_names = list(
            options['target']
            if options['target'] else sorted(graph_factories.keys())
        )

        # Create a graph with all requested subgraphs
        graph = bonobo.Graph()
        for name in graph_names:
            graph = graph_factories[name](graph, **graph_options)

        bonobo.run(graph, services=get_services())

        if options['sync']:
            # TODO: when parallel option for node will be implemented, need to be rewriten to use a graph.
            import boto3

            s3 = boto3.client('s3')

            local_dir = get_datasets_dir()
            for root, dirs, files in os.walk(local_dir):
                for filename in files:
                    local_path = os.path.join(root, filename)
                    relative_path = os.path.relpath(local_path, local_dir)
                    s3_path = os.path.join(
                        get_minor_version(), relative_path
                    )

                    try:
                        s3.head_object(
                            Bucket='bonobo-examples', Key=s3_path
                        )
                    except:
                        s3.upload_file(
                            local_path,
                            'bonobo-examples',
                            s3_path,
                            ExtraArgs={
                                'ACL': 'public-read'
                            }
                        )



# -*- coding: utf-8 -*-
#
# Copyright 2012-2014 Romain Dorgueil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta, abstractmethod
from queue import Queue

from bonobo.constants import BEGIN, END
from bonobo.errors import AbstractError, InactiveReadableError, InactiveWritableError
from bonobo.nodes import noop

BUFFER_SIZE = 8192


class Readable(metaclass=ABCMeta):
    """Interface for things you can read from."""

    @abstractmethod
    def get(self, block=True, timeout=None):
        """Read. Block/timeout are there for Queue compat."""
        raise AbstractError(self.get)


class Writable(metaclass=ABCMeta):
    """Interface for things you can write to."""

    @abstractmethod
    def put(self, data, block=True, timeout=None):
        """Write. Block/timeout are there for Queue compat."""
        raise AbstractError(self.put)


class Input(Queue, Readable, Writable):
    def __init__(self, maxsize=BUFFER_SIZE):
        Queue.__init__(self, maxsize)

        self._runlevel = 0
        self._writable_runlevel = 0
        self.on_initialize = noop
        self.on_begin = noop
        self.on_end = noop
        self.on_finalize = noop

    def put(self, data, block=True, timeout=None):
        # Begin token is a metadata to raise the input runlevel.
        if data == BEGIN:
            if not self._runlevel:
                self.on_initialize()

            self._runlevel += 1
            self._writable_runlevel += 1

            # callback
            self.on_begin()

            return

        # Check we are actually able to receive data.
        if self._writable_runlevel < 1:
            raise InactiveWritableError('Cannot put() on an inactive {}.'.format(Writable.__name__))

        if data == END:
            self._writable_runlevel -= 1

        return Queue.put(self, data, block, timeout)

    def _decrement_runlevel(self):
        if self._runlevel == 1:
            self.on_finalize()
        self._runlevel -= 1
        self.on_end()

    def get(self, block=True, timeout=None):
        if not self.alive:
            raise InactiveReadableError('Cannot get() on an inactive {}.'.format(Readable.__name__))

        data = Queue.get(self, block, timeout)

        if data == END:
            self._decrement_runlevel()

            if not self.alive:
                raise InactiveReadableError(
                    'Cannot get() on an inactive {} (runlevel just reached 0).'.format(Readable.__name__)
                )
            return self.get(block, timeout)

        return data

    def shutdown(self):
        while self._runlevel >= 1:
            self._decrement_runlevel()

    def empty(self):
        self.mutex.acquire()
        while self._qsize() and self.queue[0] == END:
            self._runlevel -= 1
            Queue._get(self)
        self.mutex.release()

        return Queue.empty(self)

    @property
    def alive(self):
        return self._runlevel > 0



from bonobo.nodes.basics import *
from bonobo.nodes.basics import __all__ as _all_basics
from bonobo.nodes.filter import Filter
from bonobo.nodes.io import *
from bonobo.nodes.io import __all__ as _all_io
from bonobo.nodes.throttle import RateLimited

__all__ = _all_basics + _all_io + ['Filter', 'RateLimited']



import json
from collections import OrderedDict

from bonobo.config import Method
from bonobo.config.processors import ContextProcessor, use_context
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.base import FileHandler
from bonobo.nodes.io.file import FileReader, FileWriter


class JsonHandler(FileHandler):
    eol = ',\n'
    prefix, suffix = '[', ']'


class LdjsonHandler(FileHandler):
    eol = '\n'
    prefix, suffix = '', ''


class JsonReader(JsonHandler, FileReader):
    @Method(positional=False)
    def loader(self, file):
        return json.loads(file)

    def read(self, file, *, fs):
        yield from self.loader(file.read())

    __call__ = read


class LdjsonReader(LdjsonHandler, JsonReader):
    """
    Read a stream of line-delimited JSON objects (one object per line).

    Not to be mistaken with JSON-LD (where LD stands for linked data).

    """

    def read(self, file, *, fs):
        yield from map(self.loader, file)

    __call__ = read


@use_context
class JsonWriter(JsonHandler, FileWriter):
    @ContextProcessor
    def envelope(self, context, file, *, fs):
        file.write(self.prefix)
        yield
        file.write(self.suffix)

    def write(self, file, context, *args, fs):
        """
        Write a json row on the next line of file pointed by ctx.file.

        :param ctx:
        :param row:
        """
        context.setdefault('lineno', 0)
        fields = context.get_input_fields()

        if fields:
            prefix = self.eol if context.lineno else ''
            self._write_line(file, prefix + json.dumps(OrderedDict(zip(fields, args))))
            context.lineno += 1
        else:
            for arg in args:
                prefix = self.eol if context.lineno else ''
                self._write_line(file, prefix + json.dumps(arg))
                context.lineno += 1

        return NOT_MODIFIED

    __call__ = write


@use_context
class LdjsonWriter(LdjsonHandler, JsonWriter):
    """
    Write a stream of Line-delimited JSON objects (one object per line).

    Not to be mistaken with JSON-LD (where LD stands for linked data).

    """



import inspect
import pprint
import re
import threading
import types
from contextlib import ContextDecorator

from bonobo.config.options import Option
from bonobo.errors import MissingServiceImplementationError

_service_name_re = re.compile(r"^[^\d\W]\w*(:?\.[^\d\W]\w*)*$", re.UNICODE)


def validate_service_name(name):
    if not _service_name_re.match(name):
        raise ValueError('Invalid service name {!r}.'.format(name))
    return name


class Service(Option):
    """
    A Service is a special kind of option defining a dependency to something that will be resolved at runtime, using an
    identifier. For example, you can create a Configurable that has a "database" Service in its attribute, meaning that
    you'll define which database to use, by name, when creating the instance of this class, then provide an
    implementation when running the graph using a strategy.
    
    Example::
    
        import bonobo
    
        class QueryExtractor(bonobo.Configurable):
            database = bonobo.Service(default='sqlalchemy.engine.default')
            
        graph = bonobo.Graph(
            QueryExtractor(database='sqlalchemy.engine.secondary'),
            *more_transformations,
        )
        
        if __name__ == '__main__':
            engine = create_engine('... dsn ...')
            bonobo.run(graph, services={
                'sqlalchemy.engine.secondary': engine
            })
            
    The main goal is not to tie transformations to actual dependencies, so the same can be run in different contexts
    (stages like preprod, prod, or tenants like client1, client2, or anything you want).

    .. attribute:: name

        Service name will be used to retrieve the implementation at runtime.
    
    """

    def __init__(self, name, __doc__=None):
        super().__init__(str, required=False, default=name, __doc__=__doc__)

    def __set__(self, inst, value):
        inst._options_values[self.name] = validate_service_name(value)

    def resolve(self, inst, services):
        try:
            name = getattr(inst, self.name)
        except AttributeError:
            name = self.name
        return services.get(name)


class Container(dict):
    def __new__(cls, *args, **kwargs):
        if len(args) == 1:
            assert not len(kwargs), 'only one usage at a time, my dear.'
            if not (args[0]):
                return super().__new__(cls)
            if isinstance(args[0], cls):
                return cls
        return super().__new__(cls, *args, **kwargs)

    def kwargs_for(self, mixed):
        try:
            options = dict(mixed.__options__)
        except AttributeError:
            options = {}

        return {name: option.resolve(mixed, self) for name, option in options.items() if isinstance(option, Service)}

    def get(self, name, default=None):
        if not name in self:
            if default:
                return default
            raise MissingServiceImplementationError(
                'Cannot resolve service {!r} using provided service collection.'.format(name)
            )
        value = super().get(name)
        # XXX this is not documented and can lead to errors.
        if isinstance(value, types.LambdaType):
            value = value(self)
        return value


def create_container(services=None, factory=Container):
    """
    Create a container with reasonable default service implementations for commonly use, standard-named, services.

    Services:
    - `fs` defaults to a fs2 instance based on current working directory
    - `http`defaults to requests

    :param services:
    :return:
    """
    container = factory(services) if services else factory()

    if not 'fs' in container:
        import bonobo
        container.setdefault('fs', bonobo.open_fs())

    if not 'http' in container:
        import requests
        container.setdefault('http', requests)

    return container


class Exclusive(ContextDecorator):
    """
    Decorator and context manager used to require exclusive usage of an object, most probably a service. It's usefull
    for example if call order matters on a service implementation (think of an http api that requires a nonce or version
    parameter ...).

    Usage:

        >>> def handler(some_service):
        ...     with Exclusive(some_service):
        ...         some_service.call_1()
        ...         some_service.call_2()
        ...         some_service.call_3()

    This will ensure that nobody else is using the same service while in the "with" block, using a lock primitive to
    ensure that.

    """
    _locks = {}

    def __init__(self, wrapped):
        self._wrapped = wrapped

    def get_lock(self):
        _id = id(self._wrapped)
        if not _id in Exclusive._locks:
            Exclusive._locks[_id] = threading.RLock()
        return Exclusive._locks[_id]

    def __enter__(self):
        self.get_lock().acquire()
        return self._wrapped

    def __exit__(self, *exc):
        self.get_lock().release()


def use(*service_names):
    def decorate(mixed):
        try:
            options = mixed.__options__
        except AttributeError:
            mixed.__options__ = options = {}

        for service_name in service_names:
            service = Service(service_name)
            service.name = service_name
            options[service_name] = service
        return mixed

    return decorate



import logging

import os

from bonobo.errors import ValidationError


def to_bool(s):
    if s is None:
        return False
    if type(s) is bool:
        return s
    if len(s):
        if s.lower() in ('f', 'false', 'n', 'no', '0'):
            return False
        return True
    return False


class Setting:
    __all__ = {}

    @classmethod
    def clear_all(cls):
        for setting in Setting.__all__.values():
            setting.clear()

    def __new__(cls, name, *args, **kwargs):
        Setting.__all__[name] = super().__new__(cls)
        return Setting.__all__[name]

    def __init__(self, name, default=None, validator=None, formatter=None):
        self.name = name

        if default:
            self.default = default if callable(default) else lambda: default
        else:
            self.default = lambda: None

        self.validator = validator
        self.formatter = formatter

    def __repr__(self):
        return '<Setting {}={!r}>'.format(self.name, self.get())

    def __eq__(self, other):
        return self.get() == other

    def __bool__(self):
        return bool(self.get())

    def set(self, value):
        value = self.formatter(value) if self.formatter else value
        if self.validator and not self.validator(value):
            raise ValidationError('Invalid value {!r} for setting {}.'.format(value, self.name))
        self.value = value

    def set_if_true(self, value):
        """Sets the value to true if it is actually true. May sound strange but the main usage is enforcing some
        settings from command line."""
        if value:
            self.set(True)

    def get(self):
        try:
            return self.value
        except AttributeError:
            value = os.environ.get(self.name, None)
            if value is None:
                value = self.default()
            self.set(value)
            return self.value

    def clear(self):
        try:
            del self.value
        except AttributeError:
            pass


# Debug/verbose mode.
DEBUG = Setting('DEBUG', formatter=to_bool, default=False)

# Profile mode.
PROFILE = Setting('PROFILE', formatter=to_bool, default=False)

# Quiet mode.
QUIET = Setting('QUIET', formatter=to_bool, default=False)

# Logging level.
LOGGING_LEVEL = Setting(
    'LOGGING_LEVEL',
    formatter=logging._checkLevel,
    validator=logging._checkLevel,
    default=lambda: logging.DEBUG if DEBUG.get() else logging.INFO
)

# Input/Output format for transformations
IOFORMAT_ARG0 = 'arg0'
IOFORMAT_KWARGS = 'kwargs'

IOFORMATS = {
    IOFORMAT_ARG0,
    IOFORMAT_KWARGS,
}

IOFORMAT = Setting('IOFORMAT', default=IOFORMAT_KWARGS, validator=IOFORMATS.__contains__)


def check():
    if DEBUG.get() and QUIET.get():
        raise RuntimeError('I cannot be verbose and quiet at the same time.')


clear_all = Setting.clear_all



import argparse
import logging

import mondrian
from bonobo import settings
from bonobo.commands.base import BaseCommand, BaseGraphCommand


def entrypoint(args=None):
    """
    Main callable for "bonobo" entrypoint.

    Will load commands from "bonobo.commands" entrypoints, using stevedore.

    """

    mondrian.setup(excepthook=True)
    logger = logging.getLogger()
    logger.setLevel(settings.LOGGING_LEVEL.get())

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', '-D', action='store_true')

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    commands = {}

    def register_extension(ext, commands=commands):
        try:
            parser = subparsers.add_parser(ext.name)
            if isinstance(ext.plugin, type) and issubclass(ext.plugin, BaseCommand):
                # current way, class based.
                cmd = ext.plugin()
                cmd.add_arguments(parser)
                cmd.__name__ = ext.name
                commands[ext.name] = cmd.handle
            else:
                # old school, function based.
                commands[ext.name] = ext.plugin(parser)
        except Exception:
            logger.exception('Error while loading command {}.'.format(ext.name))

    from stevedore import ExtensionManager
    mgr = ExtensionManager(namespace='bonobo.commands')
    mgr.map(register_extension)

    parsed_args = parser.parse_args(args).__dict__

    if parsed_args.pop('debug', False):
        settings.DEBUG.set(True)
        settings.LOGGING_LEVEL.set(logging.DEBUG)
        logger.setLevel(settings.LOGGING_LEVEL.get())

    logger.debug('Command: ' + parsed_args['command'] + ' Arguments: ' + repr(parsed_args))

    # Get command handler, execute, rince.
    command = commands[parsed_args.pop('command')]
    command(**parsed_args)

    return 0



class Token:
    """Factory for signal oriented queue messages or other token types."""

    def __init__(self, name):
        self.__name__ = name

    def __repr__(self):
        return '<{}>'.format(self.__name__)


BEGIN = Token('Begin')
END = Token('End')


class Flag(Token):
    must_be_first = False
    must_be_last = False
    allows_data = True


INHERIT = Flag('Inherit')
NOT_MODIFIED = Flag('NotModified')
NOT_MODIFIED.must_be_first = True
NOT_MODIFIED.must_be_last = True
NOT_MODIFIED.allows_data = False

EMPTY = tuple()

TICK_PERIOD = 0.2



import ipywidgets as widgets
from traitlets import List, Unicode


@widgets.register('bonobo-widget.bonobo')
class BonoboWidget(widgets.DOMWidget):
    _view_name = Unicode('BonoboView').tag(sync=True)
    _model_name = Unicode('BonoboModel').tag(sync=True)
    _view_module = Unicode('bonobo-jupyter').tag(sync=True)
    _model_module = Unicode('bonobo-jupyter').tag(sync=True)
    value = List().tag(sync=True)



from bonobo.config.configurables import Configurable
from bonobo.config.functools import transformation_factory
from bonobo.config.options import Method, Option
from bonobo.config.processors import ContextProcessor, use_context, use_context_processor, use_raw_input, use_no_input
from bonobo.config.services import Container, Exclusive, Service, use, create_container
from bonobo.util import deprecated_alias

requires = deprecated_alias('requires', use)

# Bonobo's Config API
__all__ = [
    'Configurable',
    'Container',
    'ContextProcessor',
    'Exclusive',
    'Method',
    'Option',
    'Service',
    'create_container',
    'requires',
    'transformation_factory',
    'use',
    'use_context',
    'use_context_processor',
    'use_no_input',
    'use_raw_input',
]



from bonobo.execution.contexts.graph import GraphExecutionContext
from bonobo.execution.contexts.node import NodeExecutionContext
from bonobo.execution.contexts.plugin import PluginExecutionContext

__all__ = [
    'GraphExecutionContext',
    'NodeExecutionContext',
    'PluginExecutionContext',
]



import functools

import itertools


def transformation_factory(f):
    @functools.wraps(f)
    def _transformation_factory(*args, **kwargs):
        retval = f(*args, **kwargs)
        retval.__name__ = f.__name__ + '({})'.format(
            ', '.join(itertools.chain(map(repr, args), ('{}={!r}'.format(k, v) for k, v in kwargs.items())))
        )
        return retval

    _transformation_factory._partial = True

    return _transformation_factory



import logging

from bonobo.contrib.jupyter.widget import BonoboWidget
from bonobo.execution import events
from bonobo.plugins import Plugin

try:
    import IPython.core.display
except ImportError as e:
    logging.exception(
        'You must install Jupyter to use the bonobo Jupyter extension. Easiest way is to install the '
        'optional "jupyter" dependencies with «pip install bonobo[jupyter]», but you can also install a '
        'specific version by yourself.'
    )


class JupyterOutputPlugin(Plugin):
    def register(self, dispatcher):
        dispatcher.add_listener(events.START, self.setup)
        dispatcher.add_listener(events.TICK, self.tick)
        dispatcher.add_listener(events.STOPPED, self.tick)

    def unregister(self, dispatcher):
        dispatcher.remove_listener(events.STOPPED, self.tick)
        dispatcher.remove_listener(events.TICK, self.tick)
        dispatcher.remove_listener(events.START, self.setup)

    def setup(self, event):
        self.widget = BonoboWidget()
        IPython.core.display.display(self.widget)

    def tick(self, event):
        self.widget.value = [event.context[i].as_dict() for i in event.context.graph.topologically_sorted_indexes]



from urllib.parse import urlencode

import requests  # todo: make this a service so we can substitute it ?

from bonobo.config import Option
from bonobo.config.processors import ContextProcessor
from bonobo.config.configurables import Configurable
from bonobo.util.objects import ValueHolder


def path_str(path):
    return path if path.startswith('/') else '/' + path


class OpenDataSoftAPI(Configurable):
    dataset = Option(str, positional=True)
    endpoint = Option(str, required=False, default='{scheme}://{netloc}{path}')
    scheme = Option(str, required=False, default='https')
    netloc = Option(str, required=False, default='data.opendatasoft.com')
    path = Option(path_str, required=False, default='/api/records/1.0/search/')
    rows = Option(int, required=False, default=500)
    limit = Option(int, required=False)
    timezone = Option(str, required=False, default='Europe/Paris')
    kwargs = Option(dict, required=False, default=dict)

    @ContextProcessor
    def compute_path(self, context):
        params = (('dataset', self.dataset), ('timezone', self.timezone)) + tuple(sorted(self.kwargs.items()))
        yield self.endpoint.format(scheme=self.scheme, netloc=self.netloc, path=self.path) + '?' + urlencode(params)

    @ContextProcessor
    def start(self, context, base_url):
        yield ValueHolder(0)

    def __call__(self, base_url, start, *args, **kwargs):
        while (not self.limit) or (self.limit > start):
            url = '{}&start={start}&rows={rows}'.format(
                base_url, start=start.value, rows=self.rows if not self.limit else min(self.rows, self.limit - start)
            )
            resp = requests.get(url)
            records = resp.json().get('records', [])

            if not len(records):
                break

            for row in records:
                yield {
                    **row.get('fields', {}),
                    'geometry': row.get('geometry', {}),
                    'recordid': row.get('recordid'),
                }

            start += self.rows


__all__ = [
    'OpenDataSoftAPI',
]



def get_name(mixed):
    try:
        return mixed.__name__
    except AttributeError:
        return type(mixed).__name__


class Wrapper:
    def __init__(self, wrapped):
        self.wrapped = wrapped

    @property
    def __name__(self):
        return getattr(self.wrapped, '__name__', getattr(type(self.wrapped), '__name__', repr(self.wrapped)))

    name = __name__


class ValueHolder:
    """
    Decorator holding a value in a given memory adress, effectively allowing to "hold" even an immutable typed value as the state of a
    node, allowing different actors to mutate it durably.

    For the sake of concistency, all operator methods have been implemented (see https://docs.python.org/3/reference/datamodel.html) or
    at least all in a certain category, but it feels like a more correct method should exist, like with a getattr-something on the
    value. Let's see later.
    
    """

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        # XXX deprecated
        return self._value

    def get(self):
        return self._value

    def set(self, new_value):
        self._value = new_value

    def __bool__(self):
        return bool(self._value)

    def __eq__(self, other):
        return self._value == other

    def __ne__(self, other):
        return self._value != other

    def __repr__(self):
        return repr(self._value)

    def __lt__(self, other):
        return self._value < other

    def __le__(self, other):
        return self._value <= other

    def __gt__(self, other):
        return self._value > other

    def __ge__(self, other):
        return self._value >= other

    def __add__(self, other):
        return self._value + other

    def __radd__(self, other):
        return other + self._value

    def __iadd__(self, other):
        self._value += other
        return self

    def __sub__(self, other):
        return self._value - other

    def __rsub__(self, other):
        return other - self._value

    def __isub__(self, other):
        self._value -= other
        return self

    def __mul__(self, other):
        return self._value * other

    def __rmul__(self, other):
        return other * self._value

    def __imul__(self, other):
        self._value *= other
        return self

    def __matmul__(self, other):
        return self._value @ other

    def __rmatmul__(self, other):
        return other @ self._value

    def __imatmul__(self, other):
        self._value @= other
        return self

    def __truediv__(self, other):
        return self._value / other

    def __rtruediv__(self, other):
        return other / self._value

    def __itruediv__(self, other):
        self._value /= other
        return self

    def __floordiv__(self, other):
        return self._value // other

    def __rfloordiv__(self, other):
        return other // self._value

    def __ifloordiv__(self, other):
        self._value //= other
        return self

    def __mod__(self, other):
        return self._value % other

    def __rmod__(self, other):
        return other % self._value

    def __imod__(self, other):
        self._value %= other
        return self

    def __divmod__(self, other):
        return divmod(self._value, other)

    def __rdivmod__(self, other):
        return divmod(other, self._value)

    def __pow__(self, other):
        return self._value**other

    def __rpow__(self, other):
        return other**self._value

    def __ipow__(self, other):
        self._value **= other
        return self

    def __lshift__(self, other):
        return self._value << other

    def __rlshift__(self, other):
        return other << self._value

    def __ilshift__(self, other):
        self._value <<= other
        return self

    def __rshift__(self, other):
        return self._value >> other

    def __rrshift__(self, other):
        return other >> self._value

    def __irshift__(self, other):
        self._value >>= other
        return self

    def __and__(self, other):
        return self._value & other

    def __rand__(self, other):
        return other & self._value

    def __iand__(self, other):
        self._value &= other
        return self

    def __xor__(self, other):
        return self._value ^ other

    def __rxor__(self, other):
        return other ^ self._value

    def __ixor__(self, other):
        self._value ^= other
        return self

    def __or__(self, other):
        return self._value | other

    def __ror__(self, other):
        return other | self._value

    def __ior__(self, other):
        self._value |= other
        return self

    def __neg__(self):
        return -self._value

    def __pos__(self):
        return +self._value

    def __abs__(self):
        return abs(self._value)

    def __invert__(self):
        return ~self._value

    def __len__(self):
        return len(self._value)

    def __contains__(self, item):
        return item in self._value

    def __getitem__(self, item):
        return self._value[item]

    def __setitem__(self, key, value):
        self._value[key] = value

    def __getattr__(self, item):
        return getattr(self._value, item)


def get_attribute_or_create(obj, attr, default):
    try:
        return getattr(obj, attr)
    except AttributeError:
        setattr(obj, attr, default)
        return getattr(obj, attr)



def create_or_update(model, *, defaults=None, save=True, **kwargs):
    """
    Create or update a django model instance.

    :param model:
    :param defaults:
    :param kwargs:
    :return: object, created, updated

    """
    obj, created = model._default_manager.get_or_create(defaults=defaults, **kwargs)

    updated = False
    if not created:
        if defaults:
            for k, v in defaults.items():
                if getattr(obj, k) != v:
                    setattr(obj, k, v)
                    updated = True

        if updated and save:
            obj.save()

    return obj, created, updated



# This file is autogenerated by medikit code generator.
# All changes will be overwritten.

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Py3 compatibility hacks, borrowed from IPython.
try:
    execfile
except NameError:

    def execfile(fname, globs, locs=None):
        locs = locs or globs
        exec(compile(open(fname).read(), fname, "exec"), globs, locs)


# Get the long description from the README file
try:
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = ''

# Get the classifiers from the classifiers file
tolines = lambda c: list(filter(None, map(lambda s: s.strip(), c.split('\n'))))
try:
    with open(path.join(here, 'classifiers.txt'), encoding='utf-8') as f:
        classifiers = tolines(f.read())
except:
    classifiers = []

version_ns = {}
try:
    execfile(path.join(here, 'bonobo/_version.py'), version_ns)
except EnvironmentError:
    version = 'dev'
else:
    version = version_ns.get('__version__', 'dev')

setup(
    author='Romain Dorgueil',
    author_email='romain@dorgueil.net',
    data_files=[(
        'share/jupyter/nbextensions/bonobo-jupyter', [
            'bonobo/contrib/jupyter/static/extension.js', 'bonobo/contrib/jupyter/static/index.js',
            'bonobo/contrib/jupyter/static/index.js.map'
        ]
    )],
    description=('Bonobo, a simple, modern and atomic extract-transform-load toolkit for '
                 'python 3.5+.'),
    license='Apache License, Version 2.0',
    name='bonobo',
    version=version,
    long_description=long_description,
    classifiers=classifiers,
    packages=find_packages(exclude=['ez_setup', 'example', 'test']),
    include_package_data=True,
    install_requires=[
        'fs (~= 2.0)', 'graphviz (>= 0.8, < 0.9)', 'jinja2 (~= 2.9)', 'mondrian (~= 0.6)', 'packaging (~= 16.0)',
        'psutil (~= 5.4)', 'python-slugify (~= 1.2.0)', 'requests (~= 2.0)', 'stevedore (~= 1.27)', 'whistle (~= 1.0)'
    ],
    extras_require={
        'dev': [
            'cookiecutter (>= 1.5, < 1.6)', 'coverage (>= 4.4, < 5.0)', 'pytest (>= 3.1, < 4.0)',
            'pytest-cov (>= 2.5, < 3.0)', 'pytest-sugar (>= 0.8, < 0.9)', 'pytest-timeout (>= 1, < 2)',
            'sphinx (>= 1.6, < 2.0)', 'sphinx-sitemap (>= 0.2, < 0.3)', 'yapf'
        ],
        'docker': ['bonobo-docker (~= 0.6.0a1)'],
        'jupyter': ['ipywidgets (~= 6.0)', 'jupyter (~= 1.0)'],
        'sqlalchemy': ['bonobo-sqlalchemy (~= 0.6.0a1)']
    },
    entry_points={
        'bonobo.commands': [
            'convert = bonobo.commands.convert:ConvertCommand', 'download = bonobo.commands.download:DownloadCommand',
            'examples = bonobo.commands.examples:ExamplesCommand', 'init = bonobo.commands.init:InitCommand',
            'inspect = bonobo.commands.inspect:InspectCommand', 'run = bonobo.commands.run:RunCommand',
            'version = bonobo.commands.version:VersionCommand'
        ],
        'console_scripts': ['bonobo = bonobo.commands:entrypoint']
    },
    url='https://www.bonobo-project.org/',
    download_url='https://github.com/python-bonobo/bonobo/tarball/{version}'.format(version=version),
)



from collections import namedtuple


def isconfigurable(mixed):
    """
    Check if the given argument is an instance of :class:`bonobo.config.Configurable`.

    :param mixed:
    :return: bool
    """
    from bonobo.config.configurables import Configurable
    return isinstance(mixed, Configurable)


def isconfigurabletype(mixed, *, strict=False):
    """
    Check if the given argument is an instance of :class:`bonobo.config.ConfigurableMeta`, meaning it has all the
    plumbery necessary to build :class:`bonobo.config.Configurable`-like instances.

    :param mixed:
    :param strict: should we consider partially configured objects?
    :return: bool
    """
    from bonobo.config.configurables import ConfigurableMeta, PartiallyConfigured

    if isinstance(mixed, ConfigurableMeta):
        return True

    if strict:
        return False

    if isinstance(mixed, PartiallyConfigured):
        return True

    if hasattr(mixed, '_partial') and mixed._partial:
        return True

    return False


def isoption(mixed):
    """
    Check if the given argument is an instance of :class:`bonobo.config.Option`.

    :param mixed:
    :return: bool
    """

    from bonobo.config.options import Option
    return isinstance(mixed, Option)


def ismethod(mixed):
    """
    Check if the given argument is an instance of :class:`bonobo.config.Method`.

    :param mixed:
    :return: bool
    """
    from bonobo.config.options import Method
    return isinstance(mixed, Method)


def iscontextprocessor(x):
    """
    Check if the given argument is an instance of :class:`bonobo.config.ContextProcessor`.

    :param mixed:
    :return: bool
    """
    from bonobo.config.processors import ContextProcessor
    return isinstance(x, ContextProcessor)


def istype(mixed):
    """
    Check if the given argument is a type object.

    :param mixed:
    :return: bool
    """
    return isinstance(mixed, type)


def isdict(mixed):
    """
    Check if the given argument is a dict.

    :param mixed:
    :return: bool
    """
    return isinstance(mixed, dict)


def istuple(mixed):
    """
    Check if the given argument is a tuple.

    :param mixed:
    :return: bool
    """
    return isinstance(mixed, tuple)


ConfigurableInspection = namedtuple(
    'ConfigurableInspection', [
        'type',
        'instance',
        'options',
        'processors',
        'partial',
    ]
)

ConfigurableInspection.__enter__ = lambda self: self
ConfigurableInspection.__exit__ = lambda *exc_details: None


def inspect_node(mixed, *, _partial=None):
    """
    If the given argument is somehow a :class:`bonobo.config.Configurable` object (either a subclass, an instance, or
    a partially configured instance), then it will return a :class:`ConfigurableInspection` namedtuple, used to inspect
    the configurable metadata (options). If you want to get the option values, you don't need this, it is only usefull
    to perform introspection on a configurable.

    If it's not looking like a configurable, it will raise a :class:`TypeError`.

    :param mixed:
    :return: ConfigurableInspection

    :raise: TypeError
    """
    if isconfigurabletype(mixed, strict=True):
        inst, typ = None, mixed
    elif isconfigurable(mixed):
        inst, typ = mixed, type(mixed)
    elif hasattr(mixed, 'func'):
        return inspect_node(mixed.func, _partial=(mixed.args, mixed.keywords))
    else:
        raise TypeError(
            'Not a Configurable, nor a Configurable instance and not even a partially configured Configurable. Check your inputs.'
        )

    return ConfigurableInspection(
        typ,
        inst,
        list(typ.__options__),
        list(typ.__processors__),
        _partial,
    )
